#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from decimal import Decimal

from storm.variables import *
from storm.expr import *
from storm.tests.helper import TestHelper


class Func1(NamedFunc):
    name = "func1"

class Func2(NamedFunc):
    name = "func2"

# Create columnN, tableN, and elemN variables.
for i in range(10):
    for name in ["column", "elem"]:
        exec("%s%d = SQLToken('%s%d')" % (name, i, name, i))
    for name in ["table"]:
        exec("%s%d = '%s %d'" % (name, i, name, i))


class TrackContext(FromExpr):
    context = None

@compile.when(TrackContext)
def compile_track_context(compile, expr, state):
    expr.context = state.context
    return ""

def track_contexts(n):
    return [TrackContext() for i in range(n)]


class ExprTest(TestHelper):

    def test_select_default(self):
        expr = Select(())
        self.assertEqual(expr.columns, ())
        self.assertEqual(expr.where, Undef)
        self.assertEqual(expr.tables, Undef)
        self.assertEqual(expr.default_tables, Undef)
        self.assertEqual(expr.order_by, Undef)
        self.assertEqual(expr.group_by, Undef)
        self.assertEqual(expr.limit, Undef)
        self.assertEqual(expr.offset, Undef)
        self.assertEqual(expr.distinct, False)

    def test_select_constructor(self):
        objects = [object() for i in range(9)]
        expr = Select(*objects)
        self.assertEqual(expr.columns, objects[0])
        self.assertEqual(expr.where, objects[1])
        self.assertEqual(expr.tables, objects[2])
        self.assertEqual(expr.default_tables, objects[3])
        self.assertEqual(expr.order_by, objects[4])
        self.assertEqual(expr.group_by, objects[5])
        self.assertEqual(expr.limit, objects[6])
        self.assertEqual(expr.offset, objects[7])
        self.assertEqual(expr.distinct, objects[8])

    def test_insert_default(self):
        expr = Insert(None)
        self.assertEqual(expr.map, None)
        self.assertEqual(expr.table, Undef)
        self.assertEqual(expr.default_table, Undef)
        self.assertEqual(expr.primary_columns, Undef)
        self.assertEqual(expr.primary_variables, Undef)

    def test_insert_constructor(self):
        objects = [object() for i in range(5)]
        expr = Insert(*objects)
        self.assertEqual(expr.map, objects[0])
        self.assertEqual(expr.table, objects[1])
        self.assertEqual(expr.default_table, objects[2])
        self.assertEqual(expr.primary_columns, objects[3])
        self.assertEqual(expr.primary_variables, objects[4])

    def test_update_default(self):
        expr = Update(None)
        self.assertEqual(expr.map, None)
        self.assertEqual(expr.where, Undef)
        self.assertEqual(expr.table, Undef)
        self.assertEqual(expr.default_table, Undef)

    def test_update_constructor(self):
        objects = [object() for i in range(4)]
        expr = Update(*objects)
        self.assertEqual(expr.map, objects[0])
        self.assertEqual(expr.where, objects[1])
        self.assertEqual(expr.table, objects[2])
        self.assertEqual(expr.default_table, objects[3])

    def test_delete_default(self):
        expr = Delete()
        self.assertEqual(expr.where, Undef)
        self.assertEqual(expr.table, Undef)

    def test_delete_constructor(self):
        objects = [object() for i in range(3)]
        expr = Delete(*objects)
        self.assertEqual(expr.where, objects[0])
        self.assertEqual(expr.table, objects[1])
        self.assertEqual(expr.default_table, objects[2])

    def test_and(self):
        expr = And(elem1, elem2, elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

    def test_or(self):
        expr = Or(elem1, elem2, elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

    def test_column_default(self):
        expr = Column()
        self.assertEqual(expr.name, Undef)
        self.assertEqual(expr.table, Undef)
        self.assertIdentical(expr.compile_cache, None)

        # Test for identity. We don't want False there.
        self.assertIs(expr.primary, 0)

        self.assertEqual(expr.variable_factory, Variable)

    def test_column_constructor(self):
        objects = [object() for i in range(3)]
        objects.insert(2, True)
        expr = Column(*objects)
        self.assertEqual(expr.name, objects[0])
        self.assertEqual(expr.table, objects[1])

        # Test for identity. We don't want True there either.
        self.assertIs(expr.primary, 1)

        self.assertEqual(expr.variable_factory, objects[3])

    def test_func(self):
        expr = Func("myfunc", elem1, elem2)
        self.assertEqual(expr.name, "myfunc")
        self.assertEqual(expr.args, (elem1, elem2))

    def test_named_func(self):
        class MyFunc(NamedFunc):
            name = "myfunc"
        expr = MyFunc(elem1, elem2)
        self.assertEqual(expr.name, "myfunc")
        self.assertEqual(expr.args, (elem1, elem2))

    def test_like(self):
        expr = Like(elem1, elem2)
        self.assertEqual(expr.expr1, elem1)
        self.assertEqual(expr.expr2, elem2)

    def test_like_escape(self):
        expr = Like(elem1, elem2, elem3)
        self.assertEqual(expr.expr1, elem1)
        self.assertEqual(expr.expr2, elem2)
        self.assertEqual(expr.escape, elem3)

    def test_like_case(self):
        expr = Like(elem1, elem2, elem3)
        self.assertEqual(expr.case_sensitive, None)
        expr = Like(elem1, elem2, elem3, True)
        self.assertEqual(expr.case_sensitive, True)
        expr = Like(elem1, elem2, elem3, False)
        self.assertEqual(expr.case_sensitive, False)

    def test_startswith(self):
        expr = Func1()
        self.assertRaises(ExprError, expr.startswith, b"not a unicode string")

        like_expr = expr.startswith("abc!!_%")
        self.assertTrue(isinstance(like_expr, Like))
        self.assertIs(like_expr.expr1, expr)
        self.assertEqual(like_expr.expr2, "abc!!!!!_!%%")
        self.assertEqual(like_expr.escape, "!")

    def test_startswith_case(self):
        expr = Func1()
        like_expr = expr.startswith("abc!!_%")
        self.assertIsNone(like_expr.case_sensitive)
        like_expr = expr.startswith("abc!!_%", case_sensitive=True)
        self.assertIs(True, like_expr.case_sensitive)
        like_expr = expr.startswith("abc!!_%", case_sensitive=False)
        self.assertIs(False, like_expr.case_sensitive)

    def test_endswith(self):
        expr = Func1()
        self.assertRaises(ExprError, expr.startswith, b"not a unicode string")

        like_expr = expr.endswith("abc!!_%")
        self.assertTrue(isinstance(like_expr, Like))
        self.assertIs(like_expr.expr1, expr)
        self.assertEqual(like_expr.expr2, "%abc!!!!!_!%")
        self.assertEqual(like_expr.escape, "!")

    def test_endswith_case(self):
        expr = Func1()
        like_expr = expr.endswith("abc!!_%")
        self.assertIsNone(like_expr.case_sensitive)
        like_expr = expr.endswith("abc!!_%", case_sensitive=True)
        self.assertIs(True, like_expr.case_sensitive)
        like_expr = expr.endswith("abc!!_%", case_sensitive=False)
        self.assertIs(False, like_expr.case_sensitive)

    def test_contains_string(self):
        expr = Func1()
        self.assertRaises(
            ExprError, expr.contains_string, b"not a unicode string")

        like_expr = expr.contains_string("abc!!_%")
        self.assertTrue(isinstance(like_expr, Like))
        self.assertIs(like_expr.expr1, expr)
        self.assertEqual(like_expr.expr2, "%abc!!!!!_!%%")
        self.assertEqual(like_expr.escape, "!")

    def test_contains_string_case(self):
        expr = Func1()
        like_expr = expr.contains_string("abc!!_%")
        self.assertIsNone(like_expr.case_sensitive)
        like_expr = expr.contains_string("abc!!_%", case_sensitive=True)
        self.assertIs(True, like_expr.case_sensitive)
        like_expr = expr.contains_string("abc!!_%", case_sensitive=False)
        self.assertIs(False, like_expr.case_sensitive)

    def test_is(self):
        expr = Is(elem1, elem2)
        self.assertEqual(expr.expr1, elem1)
        self.assertEqual(expr.expr2, elem2)

    def test_is_not(self):
        expr = IsNot(elem1, elem2)
        self.assertEqual(expr.expr1, elem1)
        self.assertEqual(expr.expr2, elem2)

    def test_eq(self):
        expr = Eq(elem1, elem2)
        self.assertEqual(expr.expr1, elem1)
        self.assertEqual(expr.expr2, elem2)

    def test_sql_default(self):
        expr = SQL(None)
        self.assertEqual(expr.expr, None)
        self.assertEqual(expr.params, Undef)
        self.assertEqual(expr.tables, Undef)

    def test_sql_constructor(self):
        objects = [object() for i in range(3)]
        expr = SQL(*objects)
        self.assertEqual(expr.expr, objects[0])
        self.assertEqual(expr.params, objects[1])
        self.assertEqual(expr.tables, objects[2])

    def test_join_expr_right(self):
        expr = JoinExpr(None)
        self.assertEqual(expr.right, None)
        self.assertEqual(expr.left, Undef)
        self.assertEqual(expr.on, Undef)

    def test_join_expr_on(self):
        on = Expr()
        expr = JoinExpr(None, on)
        self.assertEqual(expr.right, None)
        self.assertEqual(expr.left, Undef)
        self.assertEqual(expr.on, on)

    def test_join_expr_on_keyword(self):
        on = Expr()
        expr = JoinExpr(None, on=on)
        self.assertEqual(expr.right, None)
        self.assertEqual(expr.left, Undef)
        self.assertEqual(expr.on, on)

    def test_join_expr_on_invalid(self):
        on = Expr()
        self.assertRaises(ExprError, JoinExpr, None, on, None)

    def test_join_expr_right_left(self):
        objects = [object() for i in range(2)]
        expr = JoinExpr(*objects)
        self.assertEqual(expr.left, objects[0])
        self.assertEqual(expr.right, objects[1])
        self.assertEqual(expr.on, Undef)

    def test_join_expr_right_left_on(self):
        objects = [object() for i in range(3)]
        expr = JoinExpr(*objects)
        self.assertEqual(expr.left, objects[0])
        self.assertEqual(expr.right, objects[1])
        self.assertEqual(expr.on, objects[2])

    def test_join_expr_right_join(self):
        join = JoinExpr(None)
        expr = JoinExpr(None, join)
        self.assertEqual(expr.right, join)
        self.assertEqual(expr.left, None)
        self.assertEqual(expr.on, Undef)

    def test_table(self):
        objects = [object() for i in range(1)]
        expr = Table(*objects)
        self.assertEqual(expr.name, objects[0])

    def test_alias_default(self):
        expr = Alias(None)
        self.assertEqual(expr.expr, None)
        self.assertTrue(isinstance(expr.name, str))

    def test_alias_constructor(self):
        objects = [object() for i in range(2)]
        expr = Alias(*objects)
        self.assertEqual(expr.expr, objects[0])
        self.assertEqual(expr.name, objects[1])

    def test_union(self):
        expr = Union(elem1, elem2, elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

    def test_union_with_kwargs(self):
        expr = Union(elem1, elem2, all=True, order_by=(), limit=1, offset=2)
        self.assertEqual(expr.exprs, (elem1, elem2))
        self.assertEqual(expr.all, True)
        self.assertEqual(expr.order_by, ())
        self.assertEqual(expr.limit, 1)
        self.assertEqual(expr.offset, 2)

    def test_union_collapse(self):
        expr = Union(Union(elem1, elem2), elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Only first expression is collapsed.
        expr = Union(elem1, Union(elem2, elem3))
        self.assertEqual(expr.exprs[0], elem1)
        self.assertTrue(isinstance(expr.exprs[1], Union))

        # Don't collapse if all is different.
        expr = Union(Union(elem1, elem2, all=True), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Union))
        expr = Union(Union(elem1, elem2), elem3, all=True)
        self.assertTrue(isinstance(expr.exprs[0], Union))
        expr = Union(Union(elem1, elem2, all=True), elem3, all=True)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Don't collapse if limit or offset are set.
        expr = Union(Union(elem1, elem2, limit=1), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Union))
        expr = Union(Union(elem1, elem2, offset=3), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Union))

        # Don't collapse other set expressions.
        expr = Union(Except(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Except))
        expr = Union(Intersect(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))

    def test_except(self):
        expr = Except(elem1, elem2, elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

    def test_except_with_kwargs(self):
        expr = Except(elem1, elem2, all=True, order_by=(), limit=1, offset=2)
        self.assertEqual(expr.exprs, (elem1, elem2))
        self.assertEqual(expr.all, True)
        self.assertEqual(expr.order_by, ())
        self.assertEqual(expr.limit, 1)
        self.assertEqual(expr.offset, 2)

    def test_except_collapse(self):
        expr = Except(Except(elem1, elem2), elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Only first expression is collapsed.
        expr = Except(elem1, Except(elem2, elem3))
        self.assertEqual(expr.exprs[0], elem1)
        self.assertTrue(isinstance(expr.exprs[1], Except))

        # Don't collapse if all is different.
        expr = Except(Except(elem1, elem2, all=True), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Except))
        expr = Except(Except(elem1, elem2), elem3, all=True)
        self.assertTrue(isinstance(expr.exprs[0], Except))
        expr = Except(Except(elem1, elem2, all=True), elem3, all=True)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Don't collapse if limit or offset are set.
        expr = Except(Except(elem1, elem2, limit=1), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Except))
        expr = Except(Except(elem1, elem2, offset=3), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Except))

        # Don't collapse other set expressions.
        expr = Except(Union(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Union))
        expr = Except(Intersect(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))

    def test_intersect(self):
        expr = Intersect(elem1, elem2, elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

    def test_intersect_with_kwargs(self):
        expr = Intersect(
            elem1, elem2, all=True, order_by=(), limit=1, offset=2)
        self.assertEqual(expr.exprs, (elem1, elem2))
        self.assertEqual(expr.all, True)
        self.assertEqual(expr.order_by, ())
        self.assertEqual(expr.limit, 1)
        self.assertEqual(expr.offset, 2)

    def test_intersect_collapse(self):
        expr = Intersect(Intersect(elem1, elem2), elem3)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Only first expression is collapsed.
        expr = Intersect(elem1, Intersect(elem2, elem3))
        self.assertEqual(expr.exprs[0], elem1)
        self.assertTrue(isinstance(expr.exprs[1], Intersect))

        # Don't collapse if all is different.
        expr = Intersect(Intersect(elem1, elem2, all=True), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))
        expr = Intersect(Intersect(elem1, elem2), elem3, all=True)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))
        expr = Intersect(Intersect(elem1, elem2, all=True), elem3, all=True)
        self.assertEqual(expr.exprs, (elem1, elem2, elem3))

        # Don't collapse if limit or offset are set.
        expr = Intersect(Intersect(elem1, elem2, limit=1), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))
        expr = Intersect(Intersect(elem1, elem2, offset=3), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Intersect))

        # Don't collapse other set expressions.
        expr = Intersect(Union(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Union))
        expr = Intersect(Except(elem1, elem2), elem3)
        self.assertTrue(isinstance(expr.exprs[0], Except))

    def test_auto_tables(self):
        expr = AutoTables(elem1, [elem2])
        self.assertEqual(expr.expr, elem1)
        self.assertEqual(expr.tables, [elem2])

    def test_sequence(self):
        expr = Sequence(elem1)
        self.assertEqual(expr.name, elem1)


class StateTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.state = State()

    def test_attrs(self):
        self.assertEqual(self.state.parameters, [])
        self.assertEqual(self.state.auto_tables, [])
        self.assertEqual(self.state.context, None)

    def test_push_pop(self):
        self.state.parameters.extend([1, 2])
        self.state.push("parameters", [])
        self.assertEqual(self.state.parameters, [])
        self.state.pop()
        self.assertEqual(self.state.parameters, [1, 2])
        self.state.push("parameters")
        self.assertEqual(self.state.parameters, [1, 2])
        self.state.parameters.append(3)
        self.assertEqual(self.state.parameters, [1, 2, 3])
        self.state.pop()
        self.assertEqual(self.state.parameters, [1, 2])

    def test_push_pop_unexistent(self):
        self.state.push("nonexistent")
        self.assertEqual(self.state.nonexistent, None)
        self.state.nonexistent = "something"
        self.state.pop()
        self.assertEqual(self.state.nonexistent, None)


class CompileTest(TestHelper):

    def test_simple_inheritance(self):
        custom_compile = compile.create_child()
        statement = custom_compile(Func1())
        self.assertEqual(statement, "func1()")

    def test_customize(self):
        custom_compile = compile.create_child()
        @custom_compile.when(type(None))
        def compile_none(compile, state, expr):
            return "None"
        statement = custom_compile(Func1(None))
        self.assertEqual(statement, "func1(None)")

    def test_customize_inheritance(self):
        class C: pass
        compile_parent = Compile()
        compile_child = compile_parent.create_child()

        @compile_parent.when(C)
        def compile_in_parent(compile, state, expr):
            return "parent"
        statement = compile_child(C())
        self.assertEqual(statement, "parent")

        @compile_child.when(C)
        def compile_in_child(compile, state, expr):
            return "child"
        statement = compile_child(C())
        self.assertEqual(statement, "child")

    def test_precedence(self):
        e = [SQLRaw('%d' % i) for i in range(10)]
        expr = And(e[1], Or(e[2], e[3]),
                   Add(e[4], Mul(e[5], Sub(e[6], Div(e[7], Div(e[8], e[9]))))))
        statement = compile(expr)
        self.assertEqual(statement,
                         "1 AND (2 OR 3) AND 4 + 5 * (6 - 7 / (8 / 9))")

        expr = Func1(Select(Count()), [Select(Count())])
        statement = compile(expr)
        self.assertEqual(statement,
                         "func1((SELECT COUNT(*)), (SELECT COUNT(*)))")

    def test_get_precedence(self):
        self.assertTrue(compile.get_precedence(Or) <
                        compile.get_precedence(And))
        self.assertTrue(compile.get_precedence(Add) <
                        compile.get_precedence(Mul))
        self.assertTrue(compile.get_precedence(Sub) <
                        compile.get_precedence(Div))

    def test_customize_precedence(self):
        expr = And(elem1, Or(elem2, elem3))
        custom_compile = compile.create_child()
        custom_compile.set_precedence(10, And)

        custom_compile.set_precedence(11, Or)
        statement = custom_compile(expr)
        self.assertEqual(statement, "elem1 AND elem2 OR elem3")

        custom_compile.set_precedence(10, Or)
        statement = custom_compile(expr)
        self.assertEqual(statement, "elem1 AND elem2 OR elem3")

        custom_compile.set_precedence(9, Or)
        statement = custom_compile(expr)
        self.assertEqual(statement, "elem1 AND (elem2 OR elem3)")

    def test_customize_precedence_inheritance(self):
        compile_parent = compile.create_child()
        compile_child = compile_parent.create_child()

        expr = And(elem1, Or(elem2, elem3))

        compile_parent.set_precedence(10, And)

        compile_parent.set_precedence(11, Or)
        self.assertEqual(compile_child.get_precedence(Or), 11)
        self.assertEqual(compile_parent.get_precedence(Or), 11)
        statement = compile_child(expr)
        self.assertEqual(statement, "elem1 AND elem2 OR elem3")

        compile_parent.set_precedence(10, Or)
        self.assertEqual(compile_child.get_precedence(Or), 10)
        self.assertEqual(compile_parent.get_precedence(Or), 10)
        statement = compile_child(expr)
        self.assertEqual(statement, "elem1 AND elem2 OR elem3")

        compile_child.set_precedence(9, Or)
        self.assertEqual(compile_child.get_precedence(Or), 9)
        self.assertEqual(compile_parent.get_precedence(Or), 10)
        statement = compile_child(expr)
        self.assertEqual(statement, "elem1 AND (elem2 OR elem3)")

    def test_compile_sequence(self):
        expr = [elem1, Func1(), (Func2(), None)]
        statement = compile(expr)
        self.assertEqual(statement, "elem1, func1(), func2(), NULL")

    def test_compile_invalid(self):
        self.assertRaises(CompileError, compile, object())
        self.assertRaises(CompileError, compile, [object()])

    def test_bytes(self):
        state = State()
        statement = compile(b"str", state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [BytesVariable(b"str")])

    def test_unicode(self):
        state = State()
        statement = compile("str", state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [UnicodeVariable("str")])

    def test_int(self):
        state = State()
        statement = compile(1, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [IntVariable(1)])

    def test_bool(self):
        state = State()
        statement = compile(True, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [BoolVariable(1)])

    def test_float(self):
        state = State()
        statement = compile(1.1, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [FloatVariable(1.1)])

    def test_decimal(self):
        state = State()
        statement = compile(Decimal("1.1"), state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(
            state.parameters, [DecimalVariable(Decimal("1.1"))])

    def test_datetime(self):
        dt = datetime(1977, 5, 4, 12, 34)
        state = State()
        statement = compile(dt, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [DateTimeVariable(dt)])

    def test_date(self):
        d = date(1977, 5, 4)
        state = State()
        statement = compile(d, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [DateVariable(d)])

    def test_time(self):
        t = time(12, 34)
        state = State()
        statement = compile(t, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [TimeVariable(t)])

    def test_timedelta(self):
        td = timedelta(days=1, seconds=2, microseconds=3)
        state = State()
        statement = compile(td, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [TimeDeltaVariable(td)])

    def test_none(self):
        state = State()
        statement = compile(None, state)
        self.assertEqual(statement, "NULL")
        self.assertEqual(state.parameters, [])

    def test_select(self):
        expr = Select([column1, column2])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "SELECT column1, column2")
        self.assertEqual(state.parameters, [])

    def test_select_distinct(self):
        expr = Select([column1, column2], Undef, [table1], distinct=True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT DISTINCT column1, column2 FROM "table 1"')
        self.assertEqual(state.parameters, [])

    def test_select_distinct_on(self):
        expr = Select([column1, column2], Undef, [table1],
                      distinct=[column2, column1])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT DISTINCT ON (column2, column1) '
                         'column1, column2 FROM "table 1"')
        self.assertEqual(state.parameters, [])

    def test_select_where(self):
        expr = Select([column1, Func1()],
                      Func1(),
                      [table1, Func1()],
                      order_by=[column2, Func1()],
                      group_by=[column3, Func1()],
                      limit=3, offset=4)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1, func1() '
                                    'FROM "table 1", func1() '
                                    'WHERE func1() '
                                    'GROUP BY column3, func1() '
                                    'ORDER BY column2, func1() '
                                    'LIMIT 3 OFFSET 4')
        self.assertEqual(state.parameters, [])

    def test_select_join_where(self):
        expr = Select(column1,
                      Func1() == "value1",
                      Join(table1, Func2() == "value2"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 FROM '
                                    'JOIN "table 1" ON func2() = ? '
                                    'WHERE func1() = ?')
        self.assertEqual([variable.get() for variable in state.parameters],
                         ["value2", "value1"])

    def test_select_auto_table(self):
        expr = Select(Column(column1, table1),
                      Column(column2, table2) == 1),
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".column1 '
                                    'FROM "table 1", "table 2" '
                                    'WHERE "table 2".column2 = ?')
        self.assertVariablesEqual(state.parameters, [Variable(1)])

    def test_select_auto_table_duplicated(self):
        expr = Select(Column(column1, table1),
                      Column(column2, table1) == 1),
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".column1 '
                                    'FROM "table 1" WHERE '
                                    '"table 1".column2 = ?')
        self.assertVariablesEqual(state.parameters, [Variable(1)])

    def test_select_auto_table_default(self):
        expr = Select(Column(column1),
                      Column(column2) == 1,
                      default_tables=table1),
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 FROM "table 1" '
                                    'WHERE column2 = ?')
        self.assertVariablesEqual(state.parameters, [Variable(1)])

    def test_select_auto_table_default_with_joins(self):
        expr = Select(Column(column1),
                      default_tables=[table1, Join(table2)]),
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 '
                                    'FROM "table 1" JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_select_auto_table_unknown(self):
        statement = compile(Select(elem1))
        self.assertEqual(statement, "SELECT elem1")

    def test_select_auto_table_sub(self):
        col1 = Column(column1, table1)
        col2 = Column(column2, table2)
        expr = Select(col1, In(elem1, Select(col2, col1 == col2, col2.table)))
        statement = compile(expr)
        self.assertEqual(statement,
                         'SELECT "table 1".column1 FROM "table 1" WHERE '
                         'elem1 IN (SELECT "table 2".column2 FROM "table 2" '
                         'WHERE "table 1".column1 = "table 2".column2)')

    def test_select_join(self):
        expr = Select([column1, Func1()], Func1(),
                      [table1, Join(table2), Join(table3)])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1, func1() '
                                    'FROM "table 1" JOIN "table 2"'
                                    ' JOIN "table 3" '
                                    'WHERE func1()')
        self.assertEqual(state.parameters, [])

    def test_select_join_right_left(self):
        expr = Select([column1, Func1()], Func1(),
                      [table1, Join(table2, table3)])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1, func1() '
                                    'FROM "table 1", "table 2" '
                                    'JOIN "table 3" WHERE func1()')
        self.assertEqual(state.parameters, [])

    def test_select_with_strings(self):
        expr = Select(column1, "1 = 2", table1, order_by="column1",
                      group_by="column2")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 FROM "table 1" '
                                    'WHERE 1 = 2 GROUP BY column2 '
                                    'ORDER BY column1')
        self.assertEqual(state.parameters, [])

    def test_select_with_unicode(self):
        expr = Select(column1, "1 = 2", table1, order_by="column1",
                      group_by=["column2"])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 FROM "table 1" '
                                    'WHERE 1 = 2 GROUP BY column2 '
                                    'ORDER BY column1')
        self.assertEqual(state.parameters, [])

    def test_select_having(self):
        expr = Select(column1, tables=table1, order_by="column1",
                      group_by=["column2"], having="1 = 2")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT column1 FROM "table 1" '
                                    'GROUP BY column2 HAVING 1 = 2 '
                                    'ORDER BY column1')
        self.assertEqual(state.parameters, [])

    def test_select_contexts(self):
        column, where, table, order_by, group_by = track_contexts(5)
        expr = Select(column, where, table,
                      order_by=order_by, group_by=group_by)
        compile(expr)
        self.assertEqual(column.context, COLUMN)
        self.assertEqual(where.context, EXPR)
        self.assertEqual(table.context, TABLE)
        self.assertEqual(order_by.context, EXPR)
        self.assertEqual(group_by.context, EXPR)

    def test_insert(self):
        expr = Insert({column1: elem1, Func1(): Func2()}, Func2())
        state = State()
        statement = compile(expr, state)
        self.assertTrue(statement in (
                        "INSERT INTO func2() (column1, func1()) "
                        "VALUES (elem1, func2())",
                        "INSERT INTO func2() (func1(), column1) "
                        "VALUES (func2(), elem1)"), statement)
        self.assertEqual(state.parameters, [])

    def test_insert_with_columns(self):
        expr = Insert({Column(column1, table1): elem1,
                       Column(column2, table1): elem2}, table2)
        state = State()
        statement = compile(expr, state)
        self.assertTrue(statement in (
                        'INSERT INTO "table 2" (column1, column2) '
                        'VALUES (elem1, elem2)',
                        'INSERT INTO "table 2" (column2, column1) '
                        'VALUES (elem2, elem1)'), statement)
        self.assertEqual(state.parameters, [])

    def test_insert_with_columns_to_escape(self):
        expr = Insert({Column("column 1", table1): elem1}, table2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'INSERT INTO "table 2" ("column 1") VALUES (elem1)')
        self.assertEqual(state.parameters, [])

    def test_insert_with_columns_as_raw_strings(self):
        expr = Insert({"column 1": elem1}, table2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'INSERT INTO "table 2" ("column 1") VALUES (elem1)')
        self.assertEqual(state.parameters, [])

    def test_insert_auto_table(self):
        expr = Insert({Column(column1, table1): elem1})
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'INSERT INTO "table 1" (column1) '
                                    'VALUES (elem1)')
        self.assertEqual(state.parameters, [])

    def test_insert_auto_table_default(self):
        expr = Insert({Column(column1): elem1}, default_table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'INSERT INTO "table 1" (column1) '
                                    'VALUES (elem1)')
        self.assertEqual(state.parameters, [])

    def test_insert_auto_table_unknown(self):
        expr = Insert({Column(column1): elem1})
        self.assertRaises(NoTableError, compile, expr)

    def test_insert_contexts(self):
        column, value, table = track_contexts(3)
        expr = Insert({column: value}, table)
        compile(expr)
        self.assertEqual(column.context, COLUMN_NAME)
        self.assertEqual(value.context, EXPR)
        self.assertEqual(table.context, TABLE)

    def test_insert_bulk(self):
        expr = Insert((Column(column1, table1), Column(column2, table1)),
                      values=[(elem1, elem2), (elem3, elem4)])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(
            statement,
            'INSERT INTO "table 1" (column1, column2) '
            'VALUES (elem1, elem2), (elem3, elem4)')
        self.assertEqual(state.parameters, [])

    def test_insert_select(self):
        expr = Insert((Column(column1, table1), Column(column2, table1)),
                      values=Select(
                        (Column(column3, table3), Column(column4, table4))))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(
            statement,
            'INSERT INTO "table 1" (column1, column2) '
            'SELECT "table 3".column3, "table 4".column4 '
            'FROM "table 3", "table 4"')
        self.assertEqual(state.parameters, [])

    def test_update(self):
        expr = Update({column1: elem1, Func1(): Func2()}, table=Func1())
        state = State()
        statement = compile(expr, state)
        self.assertTrue(statement in (
                        "UPDATE func1() SET column1=elem1, func1()=func2()",
                        "UPDATE func1() SET func1()=func2(), column1=elem1"
                        ), statement)
        self.assertEqual(state.parameters, [])

    def test_update_with_columns(self):
        expr = Update({Column(column1, table1): elem1}, table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'UPDATE "table 1" SET column1=elem1')
        self.assertEqual(state.parameters, [])

    def test_update_with_columns_to_escape(self):
        expr = Update({Column("column x", table1): elem1}, table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'UPDATE "table 1" SET "column x"=elem1')
        self.assertEqual(state.parameters, [])

    def test_update_with_columns_as_raw_strings(self):
        expr = Update({"column 1": elem1}, table=table2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'UPDATE "table 2" SET "column 1"=elem1')
        self.assertEqual(state.parameters, [])

    def test_update_where(self):
        expr = Update({column1: elem1}, Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         "UPDATE func2() SET column1=elem1 WHERE func1()")
        self.assertEqual(state.parameters, [])

    def test_update_auto_table(self):
        expr = Update({Column(column1, table1): elem1})
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'UPDATE "table 1" SET column1=elem1')
        self.assertEqual(state.parameters, [])

    def test_update_auto_table_default(self):
        expr = Update({Column(column1): elem1}, default_table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'UPDATE "table 1" SET column1=elem1')
        self.assertEqual(state.parameters, [])

    def test_update_auto_table_unknown(self):
        expr = Update({Column(column1): elem1})
        self.assertRaises(CompileError, compile, expr)

    def test_update_with_strings(self):
        expr = Update({column1: elem1}, "1 = 2", table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'UPDATE "table 1" SET column1=elem1 WHERE 1 = 2')
        self.assertEqual(state.parameters, [])

    def test_update_contexts(self):
        set_left, set_right, where, table = track_contexts(4)
        expr = Update({set_left: set_right}, where, table)
        compile(expr)
        self.assertEqual(set_left.context, COLUMN_NAME)
        self.assertEqual(set_right.context, COLUMN_NAME)
        self.assertEqual(where.context, EXPR)
        self.assertEqual(table.context, TABLE)

    def test_delete(self):
        expr = Delete(table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'DELETE FROM "table 1"')
        self.assertEqual(state.parameters, [])

    def test_delete_where(self):
        expr = Delete(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "DELETE FROM func2() WHERE func1()")
        self.assertEqual(state.parameters, [])

    def test_delete_with_strings(self):
        expr = Delete("1 = 2", table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'DELETE FROM "table 1" WHERE 1 = 2')
        self.assertEqual(state.parameters, [])

    def test_delete_auto_table(self):
        expr = Delete(Column(column1, table1) == 1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'DELETE FROM "table 1" WHERE "table 1".column1 = ?')
        self.assertVariablesEqual(state.parameters, [Variable(1)])

    def test_delete_auto_table_default(self):
        expr = Delete(Column(column1) == 1, default_table=table1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'DELETE FROM "table 1" WHERE column1 = ?')
        self.assertVariablesEqual(state.parameters, [Variable(1)])

    def test_delete_auto_table_unknown(self):
        expr = Delete(Column(column1) == 1)
        self.assertRaises(NoTableError, compile, expr)

    def test_delete_contexts(self):
        where, table = track_contexts(2)
        expr = Delete(where, table)
        compile(expr)
        self.assertEqual(where.context, EXPR)
        self.assertEqual(table.context, TABLE)

    def test_column(self):
        expr = Column(column1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "column1")
        self.assertEqual(state.parameters, [])
        self.assertEqual(expr.compile_cache, "column1")

    def test_column_table(self):
        column = Column(column1, Func1())
        expr = Select(column)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "SELECT func1().column1 FROM func1()")
        self.assertEqual(state.parameters, [])
        self.assertEqual(column.compile_cache, "column1")

    def test_column_contexts(self):
        table, = track_contexts(1)
        expr = Column(column1, table)
        compile(expr)
        self.assertEqual(table.context, COLUMN_PREFIX)

    def test_column_with_reserved_words(self):
        expr = Select(Column("name 1", "table 1"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT "table 1"."name 1" FROM "table 1"')

    def test_row(self):
        expr = Row(column1, column2)
        statement = compile(expr)
        self.assertEqual(statement, "ROW(column1, column2)")

    def test_variable(self):
        expr = Variable("value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_is_null(self):
        expr = Is(Func1(), None)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NULL")
        self.assertEqual(state.parameters, [])

    def test_is_true(self):
        expr = Is(Func1(), True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS TRUE")
        self.assertEqual(state.parameters, [])

    def test_is_false(self):
        expr = Is(Func1(), False)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS FALSE")
        self.assertEqual(state.parameters, [])

    def test_is_invalid(self):
        expr = Is(Func1(), "x")
        self.assertRaises(CompileError, compile, expr)

    def test_is_not_null(self):
        expr = IsNot(Func1(), None)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NOT NULL")
        self.assertEqual(state.parameters, [])

    def test_is_not_true(self):
        expr = IsNot(Func1(), True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NOT TRUE")
        self.assertEqual(state.parameters, [])

    def test_is_not_false(self):
        expr = IsNot(Func1(), False)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NOT FALSE")
        self.assertEqual(state.parameters, [])

    def test_is_not_invalid(self):
        expr = IsNot(Func1(), "x")
        self.assertRaises(CompileError, compile, expr)

    def test_eq(self):
        expr = Eq(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() = func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() == "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_is_in(self):
        expr = Func1().is_in(["Hello", "World"])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IN (?, ?)")
        self.assertVariablesEqual(
            state.parameters, [Variable("Hello"), Variable("World")])

    def test_is_in_empty(self):
        expr = Func1().is_in([])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "?")
        self.assertVariablesEqual(state.parameters, [BoolVariable(False)])

    def test_is_in_expr(self):
        expr = Func1().is_in(Select(column1))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IN (SELECT column1)")
        self.assertEqual(state.parameters, [])

    def test_eq_none(self):
        expr = Func1() == None

        self.assertIsNone(expr.expr2)

        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NULL")
        self.assertEqual(state.parameters, [])

    def test_ne(self):
        expr = Ne(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() != func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() != "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() != ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_ne_none(self):
        expr = Func1() != None

        self.assertIsNone(expr.expr2)

        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IS NOT NULL")
        self.assertEqual(state.parameters, [])

    def test_gt(self):
        expr = Gt(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() > func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() > "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() > ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_ge(self):
        expr = Ge(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() >= func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() >= "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() >= ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_lt(self):
        expr = Lt(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() < func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() < "value"
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() < ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_le(self):
        expr = Le(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() <= func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() <= "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() <= ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_lshift(self):
        expr = LShift(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() << func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() << "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() << ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_rshift(self):
        expr = RShift(Func1(), Func2())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() >> func2()")
        self.assertEqual(state.parameters, [])

        expr = Func1() >> "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() >> ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_like(self):
        expr = Like(Func1(), b"value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() LIKE ?")
        self.assertVariablesEqual(state.parameters, [BytesVariable(b"value")])

        expr = Func1().like("Hello")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() LIKE ?")
        self.assertVariablesEqual(state.parameters, [Variable("Hello")])

    def test_like_escape(self):
        expr = Like(Func1(), b"value", b"!")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() LIKE ? ESCAPE ?")
        self.assertVariablesEqual(state.parameters,
                          [BytesVariable(b"value"), BytesVariable(b"!")])

        expr = Func1().like("Hello", b"!")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() LIKE ? ESCAPE ?")
        self.assertVariablesEqual(state.parameters,
                          [Variable("Hello"), BytesVariable(b"!")])

    def test_like_compareable_case(self):
        expr = Func1().like("Hello")
        self.assertEqual(expr.case_sensitive, None)
        expr = Func1().like("Hello", case_sensitive=True)
        self.assertEqual(expr.case_sensitive, True)
        expr = Func1().like("Hello", case_sensitive=False)
        self.assertEqual(expr.case_sensitive, False)

    def test_in(self):
        expr = In(Func1(), b"value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IN (?)")
        self.assertVariablesEqual(state.parameters, [BytesVariable(b"value")])

        expr = In(Func1(), elem1)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() IN (elem1)")
        self.assertEqual(state.parameters, [])

    def test_and(self):
        expr = And(elem1, elem2, And(elem3, elem4))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 AND elem2 AND elem3 AND elem4")
        self.assertEqual(state.parameters, [])

        expr = Func1() & "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() AND ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_or(self):
        expr = Or(elem1, elem2, Or(elem3, elem4))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 OR elem2 OR elem3 OR elem4")
        self.assertEqual(state.parameters, [])

        expr = Func1() | "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() OR ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_and_with_strings(self):
        expr = And("elem1", "elem2")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 AND elem2")
        self.assertEqual(state.parameters, [])

    def test_or_with_strings(self):
        expr = Or("elem1", "elem2")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 OR elem2")
        self.assertEqual(state.parameters, [])

    def test_add(self):
        expr = Add(elem1, elem2, Add(elem3, elem4))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 + elem2 + elem3 + elem4")
        self.assertEqual(state.parameters, [])

        expr = Func1() + "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() + ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_sub(self):
        expr = Sub(elem1, Sub(elem2, elem3))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 - (elem2 - elem3)")
        self.assertEqual(state.parameters, [])

        expr = Sub(Sub(elem1, elem2), elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 - elem2 - elem3")
        self.assertVariablesEqual(state.parameters, [])

        expr = Func1() - "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() - ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_mul(self):
        expr = Mul(elem1, elem2, Mul(elem3, elem4))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 * elem2 * elem3 * elem4")
        self.assertEqual(state.parameters, [])

        expr = Func1() * "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() * ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_div(self):
        expr = Div(elem1, Div(elem2, elem3))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 / (elem2 / elem3)")
        self.assertEqual(state.parameters, [])

        expr = Div(Div(elem1, elem2), elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 / elem2 / elem3")
        self.assertEqual(state.parameters, [])

        expr = Func1() / "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() / ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_mod(self):
        expr = Mod(elem1, Mod(elem2, elem3))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 % (elem2 % elem3)")
        self.assertEqual(state.parameters, [])

        expr = Mod(Mod(elem1, elem2), elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 % elem2 % elem3")
        self.assertEqual(state.parameters, [])

        expr = Func1() % "value"
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() % ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_func(self):
        expr = Func("myfunc", elem1, Func1(elem2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "myfunc(elem1, func1(elem2))")
        self.assertEqual(state.parameters, [])

    def test_named_func(self):
        expr = Func1(elem1, Func2(elem2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1(elem1, func2(elem2))")
        self.assertEqual(state.parameters, [])

    def test_count(self):
        expr = Count(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "COUNT(func1())")
        self.assertEqual(state.parameters, [])

    def test_count_all(self):
        expr = Count()
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "COUNT(*)")
        self.assertEqual(state.parameters, [])

    def test_count_distinct(self):
        expr = Count(Func1(), distinct=True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "COUNT(DISTINCT func1())")
        self.assertEqual(state.parameters, [])

    def test_count_distinct_all(self):
        self.assertRaises(ValueError, Count, distinct=True)

    def test_cast(self):
        """
        The L{Cast} expression renders a C{CAST} function call with a
        user-defined input value and the type to cast it to.
        """
        expr = Cast(Func1(), "TEXT")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "CAST(func1() AS TEXT)")
        self.assertEqual(state.parameters, [])

    def test_max(self):
        expr = Max(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "MAX(func1())")
        self.assertEqual(state.parameters, [])

    def test_min(self):
        expr = Min(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "MIN(func1())")
        self.assertEqual(state.parameters, [])

    def test_avg(self):
        expr = Avg(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "AVG(func1())")
        self.assertEqual(state.parameters, [])

    def test_sum(self):
        expr = Sum(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "SUM(func1())")
        self.assertEqual(state.parameters, [])

    def test_lower(self):
        expr = Lower(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "LOWER(func1())")
        self.assertEqual(state.parameters, [])

        expr = Func1().lower()
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "LOWER(func1())")
        self.assertEqual(state.parameters, [])

    def test_upper(self):
        expr = Upper(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "UPPER(func1())")
        self.assertEqual(state.parameters, [])

        expr = Func1().upper()
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "UPPER(func1())")
        self.assertEqual(state.parameters, [])

    def test_coalesce(self):
        expr = Coalesce(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "COALESCE(func1())")
        self.assertEqual(state.parameters, [])

    def test_coalesce_with_many_arguments(self):
        expr = Coalesce(Func1(), Func2(), None)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "COALESCE(func1(), func2(), NULL)")
        self.assertEqual(state.parameters, [])

    def test_not(self):
        expr = Not(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NOT func1()")
        self.assertEqual(state.parameters, [])

    def test_exists(self):
        expr = Exists(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "EXISTS func1()")
        self.assertEqual(state.parameters, [])

    def test_neg(self):
        expr = Neg(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "- func1()")
        self.assertEqual(state.parameters, [])

        expr = -Func1()
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "- func1()")
        self.assertEqual(state.parameters, [])

    def test_asc(self):
        expr = Asc(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() ASC")
        self.assertEqual(state.parameters, [])

    def test_desc(self):
        expr = Desc(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() DESC")
        self.assertEqual(state.parameters, [])

    def test_asc_with_string(self):
        expr = Asc("column")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "column ASC")
        self.assertEqual(state.parameters, [])

    def test_desc_with_string(self):
        expr = Desc("column")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "column DESC")
        self.assertEqual(state.parameters, [])

    def test_sql(self):
        expr = SQL("expression")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "expression")
        self.assertEqual(state.parameters, [])

    def test_sql_params(self):
        expr = SQL("expression", ["params"])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "expression")
        self.assertEqual(state.parameters, ["params"])

    def test_sql_invalid_params(self):
        expr = SQL("expression", "not a list or tuple")
        self.assertRaises(CompileError, compile, expr)

    def test_sql_tables(self):
        expr = Select([column1, Func1()], SQL("expression", [], Func2()))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         "SELECT column1, func1() FROM func2() "
                         "WHERE expression")
        self.assertEqual(state.parameters, [])

    def test_sql_tables_with_list_or_tuple(self):
        sql = SQL("expression", [], [Func1(), Func2()])
        expr = Select(column1, sql)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         "SELECT column1 FROM func1(), func2() "
                         "WHERE expression")
        self.assertEqual(state.parameters, [])

        sql = SQL("expression", [], (Func1(), Func2()))
        expr = Select(column1, sql)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         "SELECT column1 FROM func1(), func2() "
                         "WHERE expression")
        self.assertEqual(state.parameters, [])

    def test_sql_comparison(self):
        expr = SQL("expression1") & SQL("expression2")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(expression1) AND (expression2)")
        self.assertEqual(state.parameters, [])

    def test_table(self):
        expr = Table(table1)
        self.assertIdentical(expr.compile_cache, None)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, '"table 1"')
        self.assertEqual(state.parameters, [])
        self.assertEqual(expr.compile_cache, '"table 1"')

    def test_alias(self):
        expr = Alias(Table(table1), "name")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "name")
        self.assertEqual(state.parameters, [])

    def test_alias_in_tables(self):
        expr = Select(column1, tables=Alias(Table(table1), "alias 1"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT column1 FROM "table 1" AS "alias 1"')
        self.assertEqual(state.parameters, [])

    def test_alias_in_tables_auto_name(self):
        expr = Select(column1, tables=Alias(Table(table1)))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement[:statement.rfind("_")+1],
                         'SELECT column1 FROM "table 1" AS "_')
        self.assertEqual(state.parameters, [])

    def test_alias_in_column_prefix(self):
        expr = Select(Column(column1, Alias(Table(table1), "alias 1")))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "alias 1".column1 '
                                    'FROM "table 1" AS "alias 1"')
        self.assertEqual(state.parameters, [])

    def test_alias_for_column(self):
        expr = Select(Alias(Column(column1, table1), "alias 1"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".column1 AS "alias 1" '
                                    'FROM "table 1"')
        self.assertEqual(state.parameters, [])

    def test_alias_union(self):
        union = Union(Select(elem1), Select(elem2))
        expr = Select(elem3, tables=Alias(union, "alias"))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         "SELECT elem3 FROM "
                         "((SELECT elem1) UNION (SELECT elem2)) AS alias")
        self.assertEqual(state.parameters, [])

    def test_distinct(self):
        """L{Distinct} adds a DISTINCT prefix to the given expression."""
        distinct = Distinct(Column(elem1))
        state = State()
        statement = compile(distinct, state)
        self.assertEqual(statement, "DISTINCT elem1")
        self.assertEqual(state.parameters, [])

    def test_join(self):
        expr = Join(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_join_on(self):
        expr = Join(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "JOIN func1() ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_join_on_with_string(self):
        expr = Join(Func1(), on="a = b")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "JOIN func1() ON a = b")
        self.assertEqual(state.parameters, [])

    def test_join_left_right(self):
        expr = Join(table1, table2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, '"table 1" JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_join_nested(self):
        expr = Join(table1, Join(table2, table3))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, '"table 1" JOIN '
                                    '("table 2" JOIN "table 3")')
        self.assertEqual(state.parameters, [])

    def test_join_double_nested(self):
        expr = Join(Join(table1, table2), Join(table3, table4))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, '"table 1" JOIN "table 2" JOIN '
                                    '("table 3" JOIN "table 4")')
        self.assertEqual(state.parameters, [])

    def test_join_table(self):
        expr = Join(Table(table1), Table(table2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, '"table 1" JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_join_contexts(self):
        table1, table2, on = track_contexts(3)
        expr = Join(table1, table2, on)
        compile(expr)
        self.assertEqual(table1.context, None)
        self.assertEqual(table2.context, None)
        self.assertEqual(on.context, EXPR)

    def test_left_join(self):
        expr = LeftJoin(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "LEFT JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_left_join_on(self):
        expr = LeftJoin(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "LEFT JOIN func1() ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_right_join(self):
        expr = RightJoin(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "RIGHT JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_right_join_on(self):
        expr = RightJoin(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "RIGHT JOIN func1() ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_natural_join(self):
        expr = NaturalJoin(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_natural_join_on(self):
        expr = NaturalJoin(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL JOIN func1() ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_natural_left_join(self):
        expr = NaturalLeftJoin(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL LEFT JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_natural_left_join_on(self):
        expr = NaturalLeftJoin(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL LEFT JOIN func1() "
                                    "ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_natural_right_join(self):
        expr = NaturalRightJoin(Func1())
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL RIGHT JOIN func1()")
        self.assertEqual(state.parameters, [])

    def test_natural_right_join_on(self):
        expr = NaturalRightJoin(Func1(), Func2() == "value")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "NATURAL RIGHT JOIN func1() "
                                    "ON func2() = ?")
        self.assertVariablesEqual(state.parameters, [Variable("value")])

    def test_union(self):
        expr = Union(Func1(), elem2, elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() UNION elem2 UNION elem3")
        self.assertEqual(state.parameters, [])

    def test_union_all(self):
        expr = Union(Func1(), elem2, elem3, all=True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() UNION ALL elem2 UNION ALL elem3")
        self.assertEqual(state.parameters, [])

    def test_union_order_by_limit_offset(self):
        expr = Union(elem1, elem2, order_by=Func1(), limit=1, offset=2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 UNION elem2 ORDER BY func1() "
                                    "LIMIT 1 OFFSET 2")
        self.assertEqual(state.parameters, [])

    def test_union_select(self):
        expr = Union(Select(elem1), Select(elem2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(SELECT elem1) UNION (SELECT elem2)")
        self.assertEqual(state.parameters, [])

    def test_union_select_nested(self):
        expr = Union(Select(elem1), Union(Select(elem2), Select(elem3)))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(SELECT elem1) UNION"
                                    " ((SELECT elem2) UNION (SELECT elem3))")
        self.assertEqual(state.parameters, [])

    def test_union_order_by_and_select(self):
        """
        When ORDER BY is present, databases usually have trouble using
        fully qualified column names.  Because of that, we transform
        pure column names into aliases, and use them in the ORDER BY.
        """
        Alias.auto_counter = 0
        column1 = Column(elem1)
        column2 = Column(elem2)
        expr = Union(Select(column1), Select(column2),
                     order_by=(column1, column2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(
            statement,
            '(SELECT elem1 AS "_1") UNION (SELECT elem2 AS "_2") '
            'ORDER BY "_1", "_2"')
        self.assertEqual(state.parameters, [])

    def test_union_contexts(self):
        select1, select2, order_by = track_contexts(3)
        expr = Union(select1, select2, order_by=order_by)
        compile(expr)
        self.assertEqual(select1.context, SELECT)
        self.assertEqual(select2.context, SELECT)
        self.assertEqual(order_by.context, COLUMN_NAME)

    def test_except(self):
        expr = Except(Func1(), elem2, elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() EXCEPT elem2 EXCEPT elem3")
        self.assertEqual(state.parameters, [])

    def test_except_all(self):
        expr = Except(Func1(), elem2, elem3, all=True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() EXCEPT ALL elem2 "
                                    "EXCEPT ALL elem3")
        self.assertEqual(state.parameters, [])

    def test_except_order_by_limit_offset(self):
        expr = Except(elem1, elem2, order_by=Func1(), limit=1, offset=2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 EXCEPT elem2 ORDER BY func1() "
                                    "LIMIT 1 OFFSET 2")
        self.assertEqual(state.parameters, [])

    def test_except_select(self):
        expr = Except(Select(elem1), Select(elem2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(SELECT elem1) EXCEPT (SELECT elem2)")
        self.assertEqual(state.parameters, [])

    def test_except_select_nested(self):
        expr = Except(Select(elem1), Except(Select(elem2), Select(elem3)))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(SELECT elem1) EXCEPT"
                                    " ((SELECT elem2) EXCEPT (SELECT elem3))")
        self.assertEqual(state.parameters, [])

    def test_except_contexts(self):
        select1, select2, order_by = track_contexts(3)
        expr = Except(select1, select2, order_by=order_by)
        compile(expr)
        self.assertEqual(select1.context, SELECT)
        self.assertEqual(select2.context, SELECT)
        self.assertEqual(order_by.context, COLUMN_NAME)

    def test_intersect(self):
        expr = Intersect(Func1(), elem2, elem3)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "func1() INTERSECT elem2 INTERSECT elem3")
        self.assertEqual(state.parameters, [])

    def test_intersect_all(self):
        expr = Intersect(Func1(), elem2, elem3, all=True)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(
            statement, "func1() INTERSECT ALL elem2 INTERSECT ALL elem3")
        self.assertEqual(state.parameters, [])

    def test_intersect_order_by_limit_offset(self):
        expr = Intersect(elem1, elem2, order_by=Func1(), limit=1, offset=2)
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "elem1 INTERSECT elem2 ORDER BY func1() "
                                    "LIMIT 1 OFFSET 2")
        self.assertEqual(state.parameters, [])

    def test_intersect_select(self):
        expr = Intersect(Select(elem1), Select(elem2))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "(SELECT elem1) INTERSECT (SELECT elem2)")
        self.assertEqual(state.parameters, [])

    def test_intersect_select_nested(self):
        expr = Intersect(
            Select(elem1), Intersect(Select(elem2), Select(elem3)))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(
            statement, "(SELECT elem1) INTERSECT"
                       " ((SELECT elem2) INTERSECT (SELECT elem3))")
        self.assertEqual(state.parameters, [])

    def test_intersect_contexts(self):
        select1, select2, order_by = track_contexts(3)
        expr = Intersect(select1, select2, order_by=order_by)
        compile(expr)
        self.assertEqual(select1.context, SELECT)
        self.assertEqual(select2.context, SELECT)
        self.assertEqual(order_by.context, COLUMN_NAME)

    def test_auto_table(self):
        expr = Select(AutoTables(1, [table1]))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT ? FROM "table 1"')
        self.assertVariablesEqual(state.parameters, [IntVariable(1)])

    def test_auto_tables_with_column(self):
        expr = Select(AutoTables(Column(elem1, table1), [table2]))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".elem1 '
                                    'FROM "table 1", "table 2"')
        self.assertEqual(state.parameters, [])

    def test_auto_tables_with_column_and_replace(self):
        expr = Select(AutoTables(Column(elem1, table1), [table2], replace=True))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".elem1 FROM "table 2"')
        self.assertEqual(state.parameters, [])

    def test_auto_tables_with_join(self):
        expr = Select(AutoTables(Column(elem1, table1), [LeftJoin(table2)]))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".elem1 FROM "table 1" '
                                    'LEFT JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_auto_tables_with_join_with_left_table(self):
        expr = Select(AutoTables(Column(elem1, table1),
                                 [LeftJoin(table1, table2)]))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, 'SELECT "table 1".elem1 FROM "table 1" '
                                    'LEFT JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_auto_tables_duplicated(self):
        expr = Select([AutoTables(Column(elem1, table1), [Join(table2)]),
                       AutoTables(Column(elem2, table2), [Join(table1)]),
                       AutoTables(Column(elem3, table1), [Join(table1)]),
                       AutoTables(Column(elem4, table3), [table1]),
                       AutoTables(Column(elem5, table1),
                                  [Join(table4, table5)])])
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT "table 1".elem1, "table 2".elem2, '
                         '"table 1".elem3, "table 3".elem4, "table 1".elem5 '
                         'FROM "table 3", "table 4" JOIN "table 5" JOIN '
                         '"table 1" JOIN "table 2"')
        self.assertEqual(state.parameters, [])

    def test_auto_tables_duplicated_nested(self):
        expr = Select(AutoTables(Column(elem1, table1), [Join(table2)]),
                      In(1, Select(AutoTables(Column(elem1, table1),
                                              [Join(table2)]))))
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement,
                         'SELECT "table 1".elem1 FROM "table 1" JOIN '
                         '"table 2" WHERE ? IN (SELECT "table 1".elem1 '
                         'FROM "table 1" JOIN "table 2")')
        self.assertVariablesEqual(state.parameters, [IntVariable(1)])

    def test_sql_token(self):
        expr = SQLToken("something")
        state = State()
        statement = compile(expr, state)
        self.assertEqual(statement, "something")
        self.assertEqual(state.parameters, [])

    def test_sql_token_spaces(self):
        expr = SQLToken("some thing")
        statement = compile(expr)
        self.assertEqual(statement, '"some thing"')

    def test_sql_token_quotes(self):
        expr = SQLToken("some'thing")
        statement = compile(expr)
        self.assertEqual(statement, '"some\'thing"')

    def test_sql_token_double_quotes(self):
        expr = SQLToken('some"thing')
        statement = compile(expr)
        self.assertEqual(statement, '"some""thing"')

    def test_sql_token_reserved(self):
        custom_compile = compile.create_child()
        custom_compile.add_reserved_words(["something"])
        expr = SQLToken("something")
        state = State()
        statement = custom_compile(expr, state)
        self.assertEqual(statement, '"something"')
        self.assertEqual(state.parameters, [])

    def test_sql_token_reserved_from_parent(self):
        expr = SQLToken("something")
        parent_compile = compile.create_child()
        child_compile = parent_compile.create_child()
        statement = child_compile(expr)
        self.assertEqual(statement, "something")
        parent_compile.add_reserved_words(["something"])
        statement = child_compile(expr)
        self.assertEqual(statement, '"something"')

    def test_sql_token_remove_reserved_word_on_child(self):
        expr = SQLToken("something")
        parent_compile = compile.create_child()
        parent_compile.add_reserved_words(["something"])
        child_compile = parent_compile.create_child()
        statement = child_compile(expr)
        self.assertEqual(statement, '"something"')
        child_compile.remove_reserved_words(["something"])
        statement = child_compile(expr)
        self.assertEqual(statement, "something")

    def test_is_reserved_word(self):
        parent_compile = compile.create_child()
        child_compile = parent_compile.create_child()
        self.assertEqual(child_compile.is_reserved_word("someTHING"), False)
        parent_compile.add_reserved_words(["SOMEthing"])
        self.assertEqual(child_compile.is_reserved_word("somETHing"), True)
        child_compile.remove_reserved_words(["soMETHing"])
        self.assertEqual(child_compile.is_reserved_word("somethING"), False)

    def test_sql1992_reserved_words(self):
        reserved_words = """
            absolute action add all allocate alter and any are as asc assertion
            at authorization avg begin between bit bit_length both by cascade
            cascaded case cast catalog char character char_ length
            character_length check close coalesce collate collation column
            commit connect connection constraint constraints continue convert
            corresponding count create cross current current_date current_time
            current_timestamp current_ user cursor date day deallocate dec
            decimal declare default deferrable deferred delete desc describe
            descriptor diagnostics disconnect distinct domain double drop else
            end end-exec escape except exception exec execute exists external
            extract false fetch first float for foreign found from full get
            global go goto grant group having hour identity immediate in
            indicator initially inner input insensitive insert int integer
            intersect interval into is isolation join key language last leading
            left level like local lower match max min minute module month names
            national natural nchar next no not null nullif numeric octet_length
            of on only open option or order outer output overlaps pad partial
            position precision prepare preserve primary prior privileges
            procedure public read real references relative restrict revoke
            right rollback rows schema scroll second section select session
            session_ user set size smallint some space sql sqlcode sqlerror
            sqlstate substring sum system_user table temporary then time
            timestamp timezone_ hour timezone_minute to trailing transaction
            translate translation trim true union unique unknown update upper
            usage user using value values varchar varying view when whenever
            where with work write year zone
            """.split()
        for word in reserved_words:
            self.assertEqual(compile.is_reserved_word(word), True)


class CompilePythonTest(TestHelper):

    def test_precedence(self):
        e = [SQLRaw('%d' % i) for i in range(10)]
        expr = And(e[1], Or(e[2], e[3]),
                   Add(e[4], Mul(e[5], Sub(e[6], Div(e[7], Div(e[8], e[9]))))))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr,
                         "1 and (2 or 3) and 4 + 5 * (6 - 7 / (8 / 9))")

    def test_get_precedence(self):
        self.assertTrue(compile_python.get_precedence(Or) <
                        compile_python.get_precedence(And))
        self.assertTrue(compile_python.get_precedence(Add) <
                        compile_python.get_precedence(Mul))
        self.assertTrue(compile_python.get_precedence(Sub) <
                        compile_python.get_precedence(Div))

    def test_compile_sequence(self):
        expr = [elem1, Variable(1), (Variable(2), None)]
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "elem1, _0, _1, None")
        self.assertEqual(state.parameters, [1, 2])

    def test_compile_invalid(self):
        self.assertRaises(CompileError, compile_python, object())
        self.assertRaises(CompileError, compile_python, [object()])

    def test_compile_unsupported(self):
        self.assertRaises(CompileError, compile_python, Expr())
        self.assertRaises(CompileError, compile_python, Func1())

    def test_bytes(self):
        py_expr = compile_python(b"str")
        self.assertEqual(py_expr, "b'str'")

    def test_unicode(self):
        py_expr = compile_python("str")
        self.assertEqual(py_expr, "'str'")

    def test_int(self):
        py_expr = compile_python(1)
        self.assertEqual(py_expr, "1")

    def test_bool(self):
        state = State()
        py_expr = compile_python(True, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, [True])

    def test_float(self):
        py_expr = compile_python(1.1)
        self.assertEqual(py_expr, repr(1.1))

    def test_datetime(self):
        dt = datetime(1977, 5, 4, 12, 34)
        state = State()
        py_expr = compile_python(dt, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, [dt])

    def test_date(self):
        d = date(1977, 5, 4)
        state = State()
        py_expr = compile_python(d, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, [d])

    def test_time(self):
        t = time(12, 34)
        state = State()
        py_expr = compile_python(t, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, [t])

    def test_timedelta(self):
        td = timedelta(days=1, seconds=2, microseconds=3)
        state = State()
        py_expr = compile_python(td, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, [td])

    def test_none(self):
        py_expr = compile_python(None)
        self.assertEqual(py_expr, "None")

    def test_column(self):
        expr = Column(column1)
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "get_column(_0)")
        self.assertEqual(state.parameters, [expr])

    def test_column_table(self):
        expr = Column(column1, table1)
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "get_column(_0)")
        self.assertEqual(state.parameters, [expr])

    def test_variable(self):
        expr = Variable("value")
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0")
        self.assertEqual(state.parameters, ["value"])

    def test_is_null(self):
        expr = Is(Variable(True), None)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is None")
        self.assertEqual(state.parameters, [True])

    def test_is_true(self):
        expr = Is(Variable(True), True)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is _1")
        self.assertEqual(state.parameters, [True, True])

    def test_is_false(self):
        expr = Is(Variable(True), False)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is _1")
        self.assertEqual(state.parameters, [True, False])

    def test_is_not_null(self):
        expr = IsNot(Variable(True), None)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is not None")
        self.assertEqual(state.parameters, [True])

    def test_is_not_true(self):
        expr = IsNot(Variable(True), True)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is not _1")
        self.assertEqual(state.parameters, [True, True])

    def test_is_not_false(self):
        expr = IsNot(Variable(True), False)
        state = State()
        statement = compile_python(expr, state)
        self.assertEqual(statement, "_0 is not _1")
        self.assertEqual(state.parameters, [True, False])

    def test_eq(self):
        expr = Eq(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 == _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_ne(self):
        expr = Ne(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 != _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_gt(self):
        expr = Gt(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 > _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_ge(self):
        expr = Ge(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 >= _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_lt(self):
        expr = Lt(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 < _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_le(self):
        expr = Le(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 <= _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_lshift(self):
        expr = LShift(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 << _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_rshift(self):
        expr = RShift(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 >> _1")
        self.assertEqual(state.parameters, [1, 2])

    def test_in(self):
        expr = In(Variable(1), Variable(2))
        state = State()
        py_expr = compile_python(expr, state)
        self.assertEqual(py_expr, "_0 in (_1,)")
        self.assertEqual(state.parameters, [1, 2])

    def test_and(self):
        expr = And(elem1, elem2, And(elem3, elem4))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 and elem2 and elem3 and elem4")

    def test_or(self):
        expr = Or(elem1, elem2, Or(elem3, elem4))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 or elem2 or elem3 or elem4")

    def test_add(self):
        expr = Add(elem1, elem2, Add(elem3, elem4))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 + elem2 + elem3 + elem4")

    def test_neg(self):
        expr = Neg(elem1)
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "-elem1")

    def test_sub(self):
        expr = Sub(elem1, Sub(elem2, elem3))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 - (elem2 - elem3)")

        expr = Sub(Sub(elem1, elem2), elem3)
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 - elem2 - elem3")

    def test_mul(self):
        expr = Mul(elem1, elem2, Mul(elem3, elem4))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 * elem2 * elem3 * elem4")

    def test_div(self):
        expr = Div(elem1, Div(elem2, elem3))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 / (elem2 / elem3)")

        expr = Div(Div(elem1, elem2), elem3)
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 / elem2 / elem3")

    def test_mod(self):
        expr = Mod(elem1, Mod(elem2, elem3))
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 % (elem2 % elem3)")

        expr = Mod(Mod(elem1, elem2), elem3)
        py_expr = compile_python(expr)
        self.assertEqual(py_expr, "elem1 % elem2 % elem3")

    def test_match(self):
        col1 = Column(column1)
        col2 = Column(column2)

        match = compile_python.get_matcher((col1 > 10) & (col2 < 10))

        self.assertTrue(match({col1: 15, col2: 5}.get))
        self.assertFalse(match({col1: 5, col2: 15}.get))

    def test_match_bad_repr(self):
        """The get_matcher() works for expressions containing values
        whose repr is not valid Python syntax."""
        class BadRepr:
            def __repr__(self):
                return "$Not a valid Python expression$"

        value = BadRepr()
        col1 = Column(column1)
        match = compile_python.get_matcher(col1 == Variable(value))
        self.assertTrue(match({col1: value}.get))


class LazyValueExprTest(TestHelper):

    def test_expr_is_lazy_value(self):
        marker = object()
        expr = SQL("Hullah!")
        variable = Variable()
        variable.set(expr)
        self.assertIs(variable.get(marker), marker)



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os


# Use a 4-digit version for development versions with 99 as the final version.
# For example, if 0.15 is the currently released version of Storm, the
# development version should be version 0.15.0.99.  This will make it obvious
# that this isn't the 0.15 release version while also allowing us to release
# an 0.15.1 if need be.  Release versions should use 2-digit version numbers,
# with 0.16 being the next release version in this example.
version = "1.0"
version_info = tuple([int(x) for x in version.split(".")])


class UndefType:

    def __repr__(self):
        return "Undef"

    def __reduce__(self):
        return "Undef"


Undef = UndefType()


# C extensions are enabled by default.  They are not used if the
# STORM_CEXTENSIONS environment variable is set to '0'.  If they can't be
# imported Storm will automatically use Python versions of the optimized code
# in the C extension.
has_cextensions = False
if os.environ.get("STORM_CEXTENSIONS") != "0":
    try:
        from storm import cextensions
        has_cextensions = True
    except ImportError as e:
        if "cextensions" not in str(e):
           raise



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from storm.event import EventSystem
from storm.tests.helper import TestHelper


class Marker:
    def __eq__(self, other):
        return self is other

    def __lt__(self, other):
        return False if self is other else NotImplemented

    __gt__ = __lt__

    def __le__(self, other):
        return True if self is other else NotImplemented

    __ge__ = __le__

marker = Marker()


class EventTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.event = EventSystem(marker)

    def test_hook_unhook_emit(self):
        called1 = []
        called2 = []
        def callback1(owner, arg1, arg2):
            called1.append((owner, arg1, arg2))
        def callback2(owner, arg1, arg2, data1, data2):
            called2.append((owner, arg1, arg2, data1, data2))

        self.event.hook("one", callback1)
        self.event.hook("one", callback1)
        self.event.hook("one", callback2, 10, 20)
        self.event.hook("two", callback2, 10, 20)
        self.event.hook("two", callback2, 10, 20)
        self.event.hook("two", callback2, 30, 40)
        self.event.hook("three", callback1)

        self.event.emit("one", 1, 2)
        self.event.emit("two", 3, 4)
        self.event.unhook("two", callback2, 10, 20)
        self.event.emit("two", 3, 4)
        self.event.emit("three", 5, 6)

        self.assertEqual(sorted(called1), [
                          (marker, 1, 2),
                          (marker, 5, 6),
                         ])
        self.assertEqual(sorted(called2), [
                          (marker, 1, 2, 10, 20),
                          (marker, 3, 4, 10, 20),
                          (marker, 3, 4, 30, 40),
                          (marker, 3, 4, 30, 40),
                         ])

    def test_unhook_by_returning_false(self):
        called = []
        def callback(owner):
            called.append(owner)
            return len(called) < 2

        self.event.hook("event", callback)

        self.event.emit("event")
        self.event.emit("event")
        self.event.emit("event")
        self.event.emit("event")

        self.assertEqual(called, [marker, marker])

    def test_weak_reference(self):
        marker = Marker()

        called = []
        def callback(owner):
            called.append(owner)

        self.event = EventSystem(marker)

        self.event.hook("event", callback)
        self.event.emit("event")

        self.assertEqual(called, [marker])
        del called[:]

        del marker
        self.event.emit("event")
        self.assertEqual(called, [])



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import shutil
import sys
import traceback

from storm.locals import StormError, Store, create_database
from storm.schema.patch import (
    Patch, PatchApplier, UnknownPatchError, BadPatchError, PatchSet)
from storm.tests.mocker import MockerTestCase


patch_test_0 = """
x = None
def apply(store):
    global x
    x = 42
    from mypackage import shared_data
    shared_data.append(42)
"""

patch_test_1 = """
y = None
def apply(store):
    global y
    y = 380
    from mypackage import shared_data
    shared_data.append(380)
"""

patch_explosion = """
from storm.locals import StormError

def apply(store):
    raise StormError('KABOOM!')
"""

patch_after_explosion = """
def apply(store):
    pass
"""

patch_no_args_apply = """
def apply():
    pass
"""

patch_missing_apply = """
def misnamed_apply(store):
    pass
"""

patch_name_error = """
def apply(store):
    blah # Comment
"""


class MockPatchStore:

    def __init__(self, database, patches=[]):
        self.database = database
        self.rolled_back = 0
        self.committed = 0
        self.patches = patches
        self.objs = []

    def execute(self, statement, params=None, noresult=False):
        pass

    def find(self, cls_spec, *args, **kwargs):
        return self.patches

    def rollback(self):
        self.rolled_back += 1

    def add(self, obj):
        self.objs.append(obj)

    def commit(self):
        self.committed += 1


class PatchApplierTest(MockerTestCase):

    def setUp(self):
        super().setUp()

        self.patchdir = self.makeDir()
        self.pkgdir = os.path.join(self.patchdir, "mypackage")
        os.makedirs(self.pkgdir)

        f = open(os.path.join(self.pkgdir, "__init__.py"), "w")
        f.write("shared_data = []")
        f.close()

        # Order of creation here is important to try to screw up the
        # patch ordering, as os.listdir returns in order of mtime (or
        # something).
        for pname, data in [("patch_380.py", patch_test_1),
                            ("patch_42.py", patch_test_0)]:
            self.add_module(pname, data)

        sys.path.append(self.patchdir)

        self.filename = self.makeFile()
        self.uri = "sqlite:///%s" % self.filename
        self.store = Store(create_database(self.uri))

        self.store.execute("CREATE TABLE patch "
                           "(version INTEGER NOT NULL PRIMARY KEY)")

        self.assertFalse(self.store.get(Patch, (42)))
        self.assertFalse(self.store.get(Patch, (380)))

        import mypackage
        self.mypackage = mypackage
        self.patch_set = PatchSet(mypackage)

        # Create another connection just to keep track of the state of the
        # whole transaction manager.  See the assertion functions below.
        self.another_store = Store(create_database("sqlite:"))
        self.another_store.execute("CREATE TABLE test (id INT)")
        self.another_store.commit()
        self.prepare_for_transaction_check()

        class Committer:

            def commit(committer):
                self.store.commit()
                self.another_store.commit()

            def rollback(committer):
                self.store.rollback()
                self.another_store.rollback()

        self.committer = Committer()
        self.patch_applier = PatchApplier(self.store, self.patch_set,
                                          self.committer)

    def tearDown(self):
        super().tearDown()
        self.committer.rollback()
        sys.path.remove(self.patchdir)
        for name in list(sys.modules):
            if name == "mypackage" or name.startswith("mypackage."):
                del sys.modules[name]

    def add_module(self, module_filename, contents):
        filename = os.path.join(self.pkgdir, module_filename)
        file = open(filename, "w")
        file.write(contents)
        file.close()

    def remove_all_modules(self):
        for filename in os.listdir(self.pkgdir):
            path = os.path.join(self.pkgdir, filename)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)

    def prepare_for_transaction_check(self):
        self.another_store.execute("DELETE FROM test")
        self.another_store.execute("INSERT INTO test VALUES (1)")

    def assert_transaction_committed(self):
        self.another_store.rollback()
        result = self.another_store.execute("SELECT * FROM test").get_one()
        self.assertEqual(result, (1,),
                         "Transaction manager wasn't committed.")

    def assert_transaction_aborted(self):
        self.another_store.commit()
        result = self.another_store.execute("SELECT * FROM test").get_one()
        self.assertEqual(result, None,
                         "Transaction manager wasn't aborted.")

    def test_apply(self):
        """
        L{PatchApplier.apply} executes the patch with the given version.
        """
        self.patch_applier.apply(42)

        x = getattr(self.mypackage, "patch_42").x
        self.assertEqual(x, 42)
        self.assertTrue(self.store.get(Patch, (42)))
        self.assertTrue("mypackage.patch_42" in sys.modules)

        self.assert_transaction_committed()

    def test_apply_with_patch_directory(self):
        """
        If the given L{PatchSet} uses sub-level patches, then the
        L{PatchApplier.apply} method will look at the per-patch directory and
        apply the relevant sub-level patch.
        """
        path = os.path.join(self.pkgdir, "patch_99")
        self.makeDir(path=path)
        self.makeFile(content="", path=os.path.join(path, "__init__.py"))
        self.makeFile(content=patch_test_0, path=os.path.join(path, "foo.py"))
        self.patch_set._sub_level = "foo"
        self.add_module("patch_99/foo.py", patch_test_0)
        self.patch_applier.apply(99)
        self.assertTrue(self.store.get(Patch, (99)))

    def test_apply_all(self):
        """
        L{PatchApplier.apply_all} executes all unapplied patches.
        """
        self.patch_applier.apply_all()

        self.assertTrue("mypackage.patch_42" in sys.modules)
        self.assertTrue("mypackage.patch_380" in sys.modules)

        x = getattr(self.mypackage, "patch_42").x
        y = getattr(self.mypackage, "patch_380").y

        self.assertEqual(x, 42)
        self.assertEqual(y, 380)

        self.assert_transaction_committed()

    def test_apply_exploding_patch(self):
        """
        L{PatchApplier.apply} aborts the transaction if the patch fails.
        """
        self.remove_all_modules()
        self.add_module("patch_666.py", patch_explosion)
        self.assertRaises(StormError, self.patch_applier.apply, 666)

        self.assert_transaction_aborted()

    def test_wb_apply_all_exploding_patch(self):
        """
        When a patch explodes the store is rolled back to make sure
        that any changes the patch made to the database are removed.
        Any other patches that have been applied successfully before
        it should not be rolled back.  Any patches pending after the
        exploding patch should remain unapplied.
        """
        self.add_module("patch_666.py", patch_explosion)
        self.add_module("patch_667.py", patch_after_explosion)
        self.assertEqual(list(self.patch_applier.get_unapplied_versions()),
                         [42, 380, 666, 667])
        self.assertRaises(StormError, self.patch_applier.apply_all)
        self.assertEqual(list(self.patch_applier.get_unapplied_versions()),
                         [666, 667])

    def test_mark_applied(self):
        """
        L{PatchApplier.mark} marks a patch has applied by inserting a new row
        in the patch table.
        """
        self.patch_applier.mark_applied(42)

        self.assertFalse("mypackage.patch_42" in sys.modules)
        self.assertFalse("mypackage.patch_380" in sys.modules)

        self.assertTrue(self.store.get(Patch, 42))
        self.assertFalse(self.store.get(Patch, 380))

        self.assert_transaction_committed()

    def test_mark_applied_all(self):
        """
        L{PatchApplier.mark_applied_all} marks all pending patches as applied.
        """
        self.patch_applier.mark_applied_all()

        self.assertFalse("mypackage.patch_42" in sys.modules)
        self.assertFalse("mypackage.patch_380" in sys.modules)

        self.assertTrue(self.store.get(Patch, 42))
        self.assertTrue(self.store.get(Patch, 380))

        self.assert_transaction_committed()

    def test_application_order(self):
        """
        L{PatchApplier.apply_all} applies the patches in increasing version
        order.
        """
        self.patch_applier.apply_all()
        self.assertEqual(self.mypackage.shared_data,
                         [42, 380])

    def test_has_pending_patches(self):
        """
        L{PatchApplier.has_pending_patches} returns C{True} if there are
        patches to be applied, C{False} otherwise.
        """
        self.assertTrue(self.patch_applier.has_pending_patches())
        self.patch_applier.apply_all()
        self.assertFalse(self.patch_applier.has_pending_patches())

    def test_abort_if_unknown_patches(self):
        """
        L{PatchApplier.mark_applied} raises and error if the patch table
        contains patches without a matching file in the patch module.
        """
        self.patch_applier.mark_applied(381)
        self.assertRaises(UnknownPatchError, self.patch_applier.apply_all)

    def test_get_unknown_patch_versions(self):
        """
        L{PatchApplier.get_unknown_patch_versions} returns the versions of all
        unapplied patches.
        """
        patches = [Patch(42), Patch(380), Patch(381)]
        my_store = MockPatchStore("database", patches=patches)
        patch_applier = PatchApplier(my_store, self.mypackage)
        self.assertEqual({381},
                         patch_applier.get_unknown_patch_versions())

    def test_no_unknown_patch_versions(self):
        """
        L{PatchApplier.get_unknown_patch_versions} returns an empty set if
        no patches are unapplied.
        """
        patches = [Patch(42), Patch(380)]
        my_store = MockPatchStore("database", patches=patches)
        patch_applier = PatchApplier(my_store, self.mypackage)
        self.assertEqual(set(), patch_applier.get_unknown_patch_versions())

    def test_patch_with_incorrect_apply(self):
        """
        L{PatchApplier.apply_all} raises an error as soon as one of the patches
        to be applied fails.
        """
        self.add_module("patch_999.py", patch_no_args_apply)
        try:
            self.patch_applier.apply_all()
        except BadPatchError as e:
            self.assertTrue("mypackage/patch_999.py" in str(e))
            self.assertTrue("takes 0 positional arguments" in str(e))
            self.assertTrue("TypeError" in str(e))
        else:
            self.fail("BadPatchError not raised")

    def test_patch_with_missing_apply(self):
        """
        L{PatchApplier.apply_all} raises an error if one of the patches to
        to be applied has no 'apply' function defined.
        """
        self.add_module("patch_999.py", patch_missing_apply)
        try:
            self.patch_applier.apply_all()
        except BadPatchError as e:
            self.assertTrue("mypackage/patch_999.py" in str(e))
            self.assertTrue("no attribute" in str(e))
            self.assertTrue("AttributeError" in str(e))
        else:
            self.fail("BadPatchError not raised")

    def test_patch_with_syntax_error(self):
        """
        L{PatchApplier.apply_all} raises an error if one of the patches to
        to be applied contains a syntax error.
        """
        self.add_module("patch_999.py", "that's not python")
        try:
            self.patch_applier.apply_all()
        except BadPatchError as e:
            self.assertTrue(" 999 " in str(e))
            self.assertTrue("SyntaxError" in str(e))
        else:
            self.fail("BadPatchError not raised")

    def test_patch_error_includes_traceback(self):
        """
        The exception raised by L{PatchApplier.apply_all} when a patch fails
        include the relevant traceback from the patch.
        """
        self.add_module("patch_999.py", patch_name_error)
        try:
            self.patch_applier.apply_all()
        except BadPatchError as e:
            self.assertTrue("mypackage/patch_999.py" in str(e))
            self.assertTrue("NameError" in str(e))
            self.assertTrue("blah" in str(e))
            formatted = traceback.format_exc()
            self.assertTrue("# Comment" in formatted)
        else:
            self.fail("BadPatchError not raised")


class PatchSetTest(MockerTestCase):

    def setUp(self):
        super().setUp()
        self.sys_dir = self.makeDir()
        self.package_dir = os.path.join(self.sys_dir, "mypackage")
        os.makedirs(self.package_dir)

        self.makeFile(
            content="", dirname=self.package_dir, basename="__init__.py")

        sys.path.append(self.sys_dir)
        import mypackage
        self.patch_package = PatchSet(mypackage, sub_level="foo")

    def tearDown(self):
        super().tearDown()
        for name in list(sys.modules):
            if name == "mypackage" or name.startswith("mypackage."):
                del sys.modules[name]

    def test_get_patch_versions(self):
        """
        The C{get_patch_versions} method returns the available patch versions,
        by looking at directories named like "patch_N".
        """
        patch_1_dir = os.path.join(self.package_dir, "patch_1")
        os.makedirs(patch_1_dir)
        self.assertEqual([1], self.patch_package.get_patch_versions())

    def test_get_patch_versions_ignores_non_patch_directories(self):
        """
        The C{get_patch_versions} method ignores files or directories not
        matching the required name pattern.
        """
        random_dir = os.path.join(self.package_dir, "random")
        os.makedirs(random_dir)
        self.assertEqual([], self.patch_package.get_patch_versions())

    def test_get_patch_module(self):
        """
        The C{get_patch_module} method returns the Python module for the patch
        with the given version.
        """
        patch_1_dir = os.path.join(self.package_dir, "patch_1")
        os.makedirs(patch_1_dir)
        self.makeFile(content="", dirname=patch_1_dir, basename="__init__.py")
        self.makeFile(content="", dirname=patch_1_dir, basename="foo.py")
        patch_module = self.patch_package.get_patch_module(1)
        self.assertEqual("mypackage.patch_1.foo", patch_module.__name__)

    def test_get_patch_module_no_sub_level(self):
        """
        The C{get_patch_module} method returns a dummy patch module if no
        sub-level file exists in the patch directory for the given version.
        """
        patch_1_dir = os.path.join(self.package_dir, "patch_1")
        os.makedirs(patch_1_dir)
        patch_module = self.patch_package.get_patch_module(1)
        store = object()
        self.assertIsNone(patch_module.apply(store))



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from datetime import datetime, date, time, timedelta
import pickle
import shutil
import sys
import os

from storm.uri import URI
from storm.expr import Select, Column, SQLToken, SQLRaw, Count, Alias
from storm.variables import (Variable, PickleVariable, BytesVariable,
                             DecimalVariable, DateTimeVariable, DateVariable,
                             TimeVariable, TimeDeltaVariable)
from storm.database import *
from storm.xid import Xid
from storm.event import EventSystem
from storm.exceptions import (
    DatabaseError, DatabaseModuleError, ConnectionBlockedError,
    DisconnectionError, Error, OperationalError, ProgrammingError)
from storm.tests.databases.proxy import ProxyTCPServer
from storm.tests.helper import MakePath


class Marker:
    pass

marker = Marker()


class DatabaseTest:

    supports_microseconds = True

    def setUp(self):
        super().setUp()
        self.create_database()
        self.create_connection()
        self.drop_tables()
        self.create_tables()
        self.create_sample_data()

    def tearDown(self):
        self.drop_sample_data()
        self.drop_tables()
        self.drop_connection()
        self.drop_database()
        super().tearDown()

    def create_database(self):
        raise NotImplementedError

    def create_connection(self):
        self.connection = self.database.connect()

    def create_tables(self):
        raise NotImplementedError

    def create_sample_data(self):
        self.connection.execute("INSERT INTO number VALUES (1, 2, 3)")
        self.connection.execute("INSERT INTO test VALUES (10, 'Title 10')")
        self.connection.execute("INSERT INTO test VALUES (20, 'Title 20')")
        self.connection.commit()

    def drop_sample_data(self):
        pass

    def drop_tables(self):
        for table in ["number", "test", "datetime_test", "bin_test"]:
            try:
                self.connection.execute("DROP TABLE " + table)
                self.connection.commit()
            except:
                self.connection.rollback()

    def drop_connection(self):
        self.connection.close()

    def drop_database(self):
        pass

    def test_create(self):
        self.assertTrue(isinstance(self.database, Database))

    def test_get_uri(self):
        """
        The get_uri() method returns the URI the database with created with.
        """
        uri = self.database.get_uri()
        self.assertIsNotNone(uri.scheme)

    def test_connection(self):
        self.assertTrue(isinstance(self.connection, Connection))

    def test_rollback(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.rollback()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertFalse(result.get_one())

    def test_rollback_twice(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.rollback()
        self.connection.rollback()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertFalse(result.get_one())

    def test_commit(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.commit()
        self.connection.rollback()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertTrue(result.get_one())

    def test_commit_twice(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.commit()
        self.connection.commit()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertTrue(result.get_one())

    def test_execute_result(self):
        result = self.connection.execute("SELECT 1")
        self.assertTrue(isinstance(result, Result))
        self.assertTrue(result.get_one())

    def test_execute_unicode_result(self):
        result = self.connection.execute("SELECT title FROM test")
        self.assertTrue(isinstance(result, Result))
        row = result.get_one()
        self.assertEqual(row, ("Title 10",))
        self.assertTrue(isinstance(row[0], str))

    def test_execute_params(self):
        result = self.connection.execute("SELECT one FROM number "
                                         "WHERE 1=?", (1,))
        self.assertTrue(result.get_one())
        result = self.connection.execute("SELECT one FROM number "
                                         "WHERE 1=?", (2,))
        self.assertFalse(result.get_one())

    def test_execute_empty_params(self):
        result = self.connection.execute("SELECT one FROM number", ())
        self.assertTrue(result.get_one())

    def test_execute_expression(self):
        result = self.connection.execute(Select(1))
        self.assertTrue(result.get_one(), (1,))

    def test_execute_expression_empty_params(self):
        result = self.connection.execute(Select(SQLRaw("1")))
        self.assertTrue(result.get_one(), (1,))

    def test_get_one(self):
        result = self.connection.execute("SELECT * FROM test ORDER BY id")
        self.assertEqual(result.get_one(), (10, "Title 10"))

    def test_get_all(self):
        result = self.connection.execute("SELECT * FROM test ORDER BY id")
        self.assertEqual(result.get_all(),
                         [(10, "Title 10"), (20, "Title 20")])

    def test_iter(self):
        result = self.connection.execute("SELECT * FROM test ORDER BY id")
        self.assertEqual([item for item in result],
                         [(10, "Title 10"), (20, "Title 20")])

    def test_simultaneous_iter(self):
        result1 = self.connection.execute("SELECT * FROM test "
                                          "ORDER BY id ASC")
        result2 = self.connection.execute("SELECT * FROM test "
                                          "ORDER BY id DESC")
        iter1 = iter(result1)
        iter2 = iter(result2)
        self.assertEqual(next(iter1), (10, "Title 10"))
        self.assertEqual(next(iter2), (20, "Title 20"))
        self.assertEqual(next(iter1), (20, "Title 20"))
        self.assertEqual(next(iter2), (10, "Title 10"))
        self.assertRaises(StopIteration, next, iter1)
        self.assertRaises(StopIteration, next, iter2)

    def test_get_insert_identity(self):
        result = self.connection.execute("INSERT INTO test (title) "
                                         "VALUES ('Title 30')")
        primary_key = (Column("id", SQLToken("test")),)
        primary_variables = (Variable(),)
        expr = result.get_insert_identity(primary_key, primary_variables)
        select = Select(Column("title", SQLToken("test")), expr)
        result = self.connection.execute(select)
        self.assertEqual(result.get_one(), ("Title 30",))

    def test_get_insert_identity_composed(self):
        result = self.connection.execute("INSERT INTO test (title) "
                                         "VALUES ('Title 30')")
        primary_key = (Column("id", SQLToken("test")),
                       Column("title", SQLToken("test")))
        primary_variables = (Variable(), Variable("Title 30"))
        expr = result.get_insert_identity(primary_key, primary_variables)
        select = Select(Column("title", SQLToken("test")), expr)
        result = self.connection.execute(select)
        self.assertEqual(result.get_one(), ("Title 30",))

    def test_datetime(self):
        value = datetime(1977, 4, 5, 12, 34, 56, 78)
        self.connection.execute("INSERT INTO datetime_test (dt) VALUES (?)",
                                (value,))
        result = self.connection.execute("SELECT dt FROM datetime_test")
        variable = DateTimeVariable()
        result.set_variable(variable, result.get_one()[0])
        if not self.supports_microseconds:
            value = value.replace(microsecond=0)
        self.assertEqual(variable.get(), value)

    def test_date(self):
        value = date(1977, 4, 5)
        self.connection.execute("INSERT INTO datetime_test (d) VALUES (?)",
                                (value,))
        result = self.connection.execute("SELECT d FROM datetime_test")
        variable = DateVariable()
        result.set_variable(variable, result.get_one()[0])
        self.assertEqual(variable.get(), value)

    def test_time(self):
        value = time(12, 34, 56, 78)
        self.connection.execute("INSERT INTO datetime_test (t) VALUES (?)",
                                (value,))
        result = self.connection.execute("SELECT t FROM datetime_test")
        variable = TimeVariable()
        result.set_variable(variable, result.get_one()[0])
        if not self.supports_microseconds:
            value = value.replace(microsecond=0)
        self.assertEqual(variable.get(), value)

    def test_timedelta(self):
        value = timedelta(12, 34, 56)
        self.connection.execute("INSERT INTO datetime_test (td) VALUES (?)",
                                (value,))
        result = self.connection.execute("SELECT td FROM datetime_test")
        variable = TimeDeltaVariable()
        result.set_variable(variable, result.get_one()[0])
        self.assertEqual(variable.get(), value)

    def test_pickle(self):
        value = {"a": 1, "b": 2}
        value_dump = pickle.dumps(value, -1)
        self.connection.execute("INSERT INTO bin_test (b) VALUES (?)",
                                (value_dump,))
        result = self.connection.execute("SELECT b FROM bin_test")
        variable = PickleVariable()
        result.set_variable(variable, result.get_one()[0])
        self.assertEqual(variable.get(), value)

    def test_binary(self):
        """Ensure database works with high bits and embedded zeros."""
        value = b"\xff\x00\xff\x00"
        self.connection.execute("INSERT INTO bin_test (b) VALUES (?)",
                                (value,))
        result = self.connection.execute("SELECT b FROM bin_test")
        variable = BytesVariable()
        result.set_variable(variable, result.get_one()[0])
        self.assertEqual(variable.get(), value)

    def test_binary_ascii(self):
        """Some databases like pysqlite2 may return unicode for strings."""
        self.connection.execute("INSERT INTO bin_test VALUES (10, 'Value')")
        result = self.connection.execute("SELECT b FROM bin_test")
        variable = BytesVariable()
        # If the following doesn't raise a TypeError we're good.
        result.set_variable(variable, result.get_one()[0])
        self.assertEqual(variable.get(), b"Value")

    def test_order_by_group_by(self):
        self.connection.execute("INSERT INTO test VALUES (100, 'Title 10')")
        self.connection.execute("INSERT INTO test VALUES (101, 'Title 10')")
        id = Column("id", "test")
        title = Column("title", "test")
        expr = Select(Count(id), group_by=title, order_by=Count(id))
        result = self.connection.execute(expr)
        self.assertEqual(result.get_all(), [(1,), (3,)])

    def test_set_decimal_variable_from_str_column(self):
        self.connection.execute("INSERT INTO test VALUES (40, '40.5')")
        variable = DecimalVariable()
        result = self.connection.execute("SELECT title FROM test WHERE id=40")
        result.set_variable(variable, result.get_one()[0])

    def test_get_decimal_variable_to_str_column(self):
        variable = DecimalVariable()
        variable.set("40.5", from_db=True)
        self.connection.execute("INSERT INTO test VALUES (40, ?)", (variable,))
        result = self.connection.execute("SELECT title FROM test WHERE id=40")
        self.assertEqual(result.get_one()[0], "40.5")

    def test_quoting(self):
        # FIXME "with'quote" should be in the list below, but it doesn't
        #       work because it breaks the parameter mark translation.
        for reserved_name in ["with space", 'with`"escape', "SELECT"]:
            reserved_name = SQLToken(reserved_name)
            expr = Select(reserved_name,
                          tables=Alias(Select(Alias(1, reserved_name))))
            result = self.connection.execute(expr)
            self.assertEqual(result.get_one(), (1,))

    def test_concurrent_behavior(self):
        """The default behavior should be to handle transactions in isolation.

        Data committed in one transaction shouldn't be visible to another
        running transaction before the later is committed or aborted.  If
        this isn't the case, the caching made by Storm (or by anything
        that works with data in memory, in fact) becomes a dangerous thing.

        For PostgreSQL, isolation level must be SERIALIZABLE.
        For MySQL, isolation level must be REPEATABLE READ (the default),
        and the InnoDB engine must be in use.
        For SQLite, the isolation level already is SERIALIZABLE when not
        in autocommit mode.  OTOH, PySQLite is nuts regarding transactional
        behavior, and will easily offer READ COMMITTED behavior inside a
        "transaction" (it didn't tell SQLite to open a transaction, in fact).
        """
        connection1 = self.connection
        connection2 = self.database.connect()
        try:
            result = connection1.execute("SELECT title FROM test WHERE id=10")
            self.assertEqual(result.get_one(), ("Title 10",))
            try:
                connection2.execute("UPDATE test SET title='Title 100' "
                                    "WHERE id=10")
                connection2.commit()
            except OperationalError as e:
                self.assertEqual(str(e), "database is locked") # SQLite blocks
            result = connection1.execute("SELECT title FROM test WHERE id=10")
            self.assertEqual(result.get_one(), ("Title 10",))
        finally:
            connection1.rollback()

    def test_wb_connect_sets_event_system(self):
        connection = self.database.connect(marker)
        self.assertEqual(connection._event, marker)

    def test_execute_sends_event(self):
        event = EventSystem(marker)
        calls = []
        def register_transaction(owner):
            calls.append(owner)
        event.hook("register-transaction", register_transaction)

        connection = self.database.connect(event)
        connection.execute("SELECT 1")
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0], marker)

    def from_database(self, row):
        return [int(item)+1 for item in row]

    def test_wb_result_get_one_goes_through_from_database(self):
        result = self.connection.execute("SELECT one, two FROM number")
        result.from_database = self.from_database
        self.assertEqual(result.get_one(), (2, 3))

    def test_wb_result_get_all_goes_through_from_database(self):
        result = self.connection.execute("SELECT one, two FROM number")
        result.from_database = self.from_database
        self.assertEqual(result.get_all(), [(2, 3)])

    def test_wb_result_iter_goes_through_from_database(self):
        result = self.connection.execute("SELECT one, two FROM number")
        result.from_database = self.from_database
        self.assertEqual(next(iter(result)), (2, 3))

    def test_rowcount_insert(self):
        # All supported backends support rowcount, so far.
        result = self.connection.execute(
            "INSERT INTO test VALUES (999, '999')")
        self.assertEqual(result.rowcount, 1)

    def test_rowcount_delete(self):
        # All supported backends support rowcount, so far.
        result = self.connection.execute("DELETE FROM test")
        self.assertEqual(result.rowcount, 2)

    def test_rowcount_update(self):
        # All supported backends support rowcount, so far.
        result = self.connection.execute(
            "UPDATE test SET title='whatever'")
        self.assertEqual(result.rowcount, 2)

    def test_expr_startswith(self):
        self.connection.execute("INSERT INTO test VALUES (30, '!!_%blah')")
        self.connection.execute("INSERT INTO test VALUES (40, '!!blah')")
        id = Column("id", SQLToken("test"))
        title = Column("title", SQLToken("test"))
        expr = Select(id, title.startswith("!!_%"))
        result = list(self.connection.execute(expr))
        self.assertEqual(result, [(30,)])

    def test_expr_endswith(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'blah_%!!')")
        self.connection.execute("INSERT INTO test VALUES (40, 'blah!!')")
        id = Column("id", SQLToken("test"))
        title = Column("title", SQLToken("test"))
        expr = Select(id, title.endswith("_%!!"))
        result = list(self.connection.execute(expr))
        self.assertEqual(result, [(30,)])

    def test_expr_contains_string(self):
        self.connection.execute("INSERT INTO test VALUES (30, 'blah_%!!x')")
        self.connection.execute("INSERT INTO test VALUES (40, 'blah!!x')")
        id = Column("id", SQLToken("test"))
        title = Column("title", SQLToken("test"))
        expr = Select(id, title.contains_string("_%!!"))
        result = list(self.connection.execute(expr))
        self.assertEqual(result, [(30,)])

    def test_block_access(self):
        """Access to the connection is blocked by block_access()."""
        self.connection.execute("SELECT 1")
        self.connection.block_access()
        self.assertRaises(ConnectionBlockedError,
                          self.connection.execute, "SELECT 1")
        self.assertRaises(ConnectionBlockedError, self.connection.commit)
        # Allow rolling back a blocked connection.
        self.connection.rollback()
        # Unblock the connection, allowing access again.
        self.connection.unblock_access()
        self.connection.execute("SELECT 1")

    def test_wrap_exception_subclasses(self):
        """Subclasses of the generic DB-API exception types are wrapped."""
        db_api_operational_error = getattr(
            self.database._exception_module, 'OperationalError')
        operational_error_types = [
            type(name, (db_api_operational_error,), {})
            for name in ('A', 'B')]
        for error_type in operational_error_types:
            error = error_type('error message')
            wrapped = self.database._wrap_exception(OperationalError, error)
            self.assertTrue(isinstance(wrapped, error_type))
            self.assertTrue(isinstance(wrapped, OperationalError))
            self.assertEqual(error_type.__name__, wrapped.__class__.__name__)
            self.assertEqual(('error message',), wrapped.args)


class TwoPhaseCommitTest:

    def setUp(self):
        super().setUp()
        self.create_database()
        self.create_connection()
        self.drop_tables()
        self.create_tables()

    def tearDown(self):
        self.drop_tables()
        self.drop_connection()
        super().tearDown()

    def create_database(self):
        raise NotImplementedError

    def create_connection(self):
        self.connection = self.database.connect()

    def create_tables(self):
        raise NotImplementedError

    def drop_tables(self):
        try:
            self.connection.execute("DROP TABLE test")
            self.connection.commit()
        except:
            self.connection.rollback()

    def drop_connection(self):
        self.connection.close()

    def test_begin(self):
        """
        begin() starts a transaction that can be ended with a two-phase commit.
        """
        xid = Xid(0, "foo", "bar")
        self.connection.begin(xid)
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.prepare()
        self.connection.commit()
        self.connection.rollback()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertTrue(result.get_one())

    def test_begin_inside_a_two_phase_transaction(self):
        """
        begin() can't be used if a two-phase transaction has already started.
        """
        xid1 = Xid(0, "foo", "bar")
        self.connection.begin(xid1)
        xid2 = Xid(1, "egg", "baz")
        self.assertRaises(ProgrammingError, self.connection.begin, xid2)

    def test_begin_after_commit(self):
        """
        After a two phase commit, it's possible to start a new transaction.
        """
        xid = Xid(0, "foo", "bar")
        self.connection.begin(xid)
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.commit()
        self.connection.begin(xid)
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertTrue(result.get_one())

    def test_begin_after_rollback(self):
        """
        After a tpc rollback, it's possible to start a new transaction.
        """
        xid = Xid(0, "foo", "bar")
        self.connection.begin(xid)
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.rollback()
        self.connection.begin(xid)
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertFalse(result.get_one())

    def test_prepare_outside_a_two_phase_transaction(self):
        """
        prepare() can't be used if a two-phase transaction has not begun yet.
        """
        self.assertRaises(ProgrammingError, self.connection.prepare)

    def test_rollback_after_prepare(self):
        """
        Calling rollback() after prepare() actually rolls back the changes.
        """
        xid = Xid(0, "foo", "bar")
        self.connection.begin(xid)
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.prepare()
        self.connection.rollback()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertFalse(result.get_one())

    def test_mixing_standard_and_two_phase_commits(self):
        """
        It's possible to mix standard and two phase commits across different
        transactions.
        """
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.commit()
        xid = Xid(0, "foo", "bar")
        self.connection.begin(xid)
        self.connection.execute("INSERT INTO test VALUES (40, 'Title 40')")
        self.connection.prepare()
        self.connection.commit()
        result = self.connection.execute("SELECT id FROM test "
                                         "WHERE id IN (30, 40)")
        self.assertEqual([(30,), (40,)], result.get_all())

    def test_recover_and_commit(self):
        """
        It's possible to recover and commit pending transactions that were
        prepared but not committed or rolled back.
        """
        # Prepare a transaction but leave it uncommitted
        self.connection.begin(Xid(0, "foo", "bar"))
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.prepare()

        # Setup a new connection and recover the prepared transaction
        # committing it
        connection2 = self.database.connect()
        self.addCleanup(connection2.close)
        result = connection2.execute("SELECT id FROM test WHERE id=30")
        connection2.rollback()
        self.assertFalse(result.get_one())
        [xid] = connection2.recover()
        self.assertEqual(0, xid.format_id)
        self.assertEqual("foo", xid.global_transaction_id)
        self.assertEqual("bar", xid.branch_qualifier)
        connection2.commit(xid)
        self.assertEqual([], connection2.recover())

        # Reconnect, changes are be visible
        self.connection.close()
        self.connection = self.database.connect()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertTrue(result.get_one())

    def test_recover_and_rollback(self):
        """
        It's possible to recover and rollback pending transactions that were
        prepared but not committed or rolled back.
        """
        # Prepare a transaction but leave it uncommitted
        self.connection.begin(Xid(0, "foo", "bar"))
        self.connection.execute("INSERT INTO test VALUES (30, 'Title 30')")
        self.connection.prepare()

        # Setup a new connection and recover the prepared transaction
        # rolling it back
        connection2 = self.database.connect()
        self.addCleanup(connection2.close)
        [xid] = connection2.recover()
        self.assertEqual(0, xid.format_id)
        self.assertEqual("foo", xid.global_transaction_id)
        self.assertEqual("bar", xid.branch_qualifier)
        connection2.rollback(xid)
        self.assertEqual([], connection2.recover())

        # Reconnect, changes were rolled back
        self.connection.close()
        self.connection = self.database.connect()
        result = self.connection.execute("SELECT id FROM test WHERE id=30")
        self.assertFalse(result.get_one())


class UnsupportedDatabaseTest:

    helpers = [MakePath]

    dbapi_module_names = []
    db_module_name = None

    def test_exception_when_unsupported(self):

        # Install a directory in front of the search path.
        module_dir = self.make_path()
        os.mkdir(module_dir)
        sys.path.insert(0, module_dir)

        # Copy the real module over to a new place, since the old one is
        # already using the real module, if it's available.
        db_module = __import__("storm.databases."+self.db_module_name,
                               None, None, [""])
        db_module_filename = db_module.__file__
        if db_module_filename.endswith(".pyc"):
            db_module_filename = db_module_filename[:-1]
        shutil.copyfile(db_module_filename,
                        os.path.join(module_dir, "_fake_.py"))

        dbapi_modules = {}
        for dbapi_module_name in self.dbapi_module_names:

            # If the real module is available, remove it from sys.modules.
            dbapi_module = sys.modules.pop(dbapi_module_name, None)
            if dbapi_module is not None:
                dbapi_modules[dbapi_module_name] = dbapi_module

            # Create a module which raises ImportError when imported, to fake
            # a missing module.
            dirname = self.make_path(path=os.path.join(module_dir,
                                                       dbapi_module_name))
            os.mkdir(dirname)
            self.make_path("raise ImportError",
                           os.path.join(module_dir, dbapi_module_name,
                                        "__init__.py"))

        # Finally, test it.
        import _fake_
        uri = URI("_fake_://db")

        try:
            self.assertRaises(DatabaseModuleError,
                              _fake_.create_from_uri, uri)
        finally:
            # Unhack the environment.
            del sys.path[0]
            del sys.modules["_fake_"]

            sys.modules.update(dbapi_modules)


class DatabaseDisconnectionMixin:

    environment_variable = ""
    host_environment_variable = ""
    default_port = None

    def setUp(self):
        super().setUp()
        self.create_database_and_proxy()
        self.create_connection()

    def tearDown(self):
        self.drop_connection()
        self.drop_database()
        self.proxy.close()
        super().tearDown()

    def is_supported(self):
        return bool(self.get_uri())

    def get_uri(self):
        """Return URI instance with a defined host (and port, for TCP)."""
        if not self.environment_variable:
            raise RuntimeError(
                "Define at least %s.environment_variable" %
                type(self).__name__)
        uri_str = os.environ.get(self.host_environment_variable)
        if uri_str:
            uri = URI(uri_str)
            if not uri.host:
                raise RuntimeError("The URI in %s must include a host." %
                                   self.host_environment_variable)
            if not uri.host.startswith("/") and not uri.port:
                if not self.default_port:
                    raise RuntimeError(
                        "Define at least %s.default_port" % type(self).__name)
                uri.port = self.default_port
            return uri
        else:
            uri_str = os.environ.get(self.environment_variable)
            if uri_str:
                uri = URI(uri_str)
                if uri.host:
                    if not uri.host.startswith("/") and not uri.port:
                        if not self.default_port:
                            raise RuntimeError(
                                "Define at least %s.default_port" %
                                type(self).__name)
                        uri.port = self.default_port
                    return uri
        return None

    def create_proxy(self, uri):
        """Create a TCP proxy forwarding requests to `uri`."""
        return ProxyTCPServer((uri.host, uri.port))

    def create_database_and_proxy(self):
        """Set up the TCP proxy and database object.

        The TCP proxy should forward requests on to the database.  The
        database object should point at the TCP proxy.
        """
        uri = self.get_uri()
        self.proxy = self.create_proxy(uri)
        uri.host, uri.port = self.proxy.server_address
        self.proxy_uri = uri
        self.database = create_database(uri)

    def create_connection(self):
        self.connection = self.database.connect()

    def drop_connection(self):
        self.connection.close()

    def drop_database(self):
        pass


class DatabaseDisconnectionTest(DatabaseDisconnectionMixin):

    def test_proxy_works(self):
        """Ensure that we can talk to the database through the proxy."""
        result = self.connection.execute("SELECT 1")
        self.assertEqual(result.get_one(), (1,))

    def test_catch_disconnect_on_execute(self):
        """Test that database disconnections get caught on execute()."""
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")

    def test_catch_disconnect_on_commit(self):
        """Test that database disconnections get caught on commit()."""
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        self.assertRaises(DisconnectionError, self.connection.commit)

    def test_wb_catch_already_disconnected_on_rollback(self):
        """Connection.rollback() swallows disconnection errors.

        If the connection is being used outside of Storm's control,
        then it is possible that Storm won't see the disconnection.
        It should be able to recover from this situation though.
        """
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        # Perform an action that should result in a disconnection.
        try:
            cursor = self.connection._raw_connection.cursor()
            cursor.execute("SELECT 1")
            cursor.fetchone()
        except Error as exc:
            self.assertTrue(self.connection.is_disconnection_error(exc))
        else:
            self.fail("Disconnection was not caught.")

        # Make sure our raw connection's rollback() raises a disconnection
        # error when called.
        try:
            self.connection._raw_connection.rollback()
        except Error as exc:
            self.assertTrue(self.connection.is_disconnection_error(exc))
        else:
            self.fail("Disconnection was not raised.")

        # Our rollback() will catch and swallow that disconnection error,
        # though.
        self.connection.rollback()

    def test_wb_catch_already_disconnected(self):
        """Storm detects connections that have already been disconnected.

        If the connection is being used outside of Storm's control,
        then it is possible that Storm won't see the disconnection.
        It should be able to recover from this situation though.
        """
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        # Perform an action that should result in a disconnection.
        try:
            cursor = self.connection._raw_connection.cursor()
            cursor.execute("SELECT 1")
            cursor.fetchone()
        except DatabaseError as exc:
            self.assertTrue(self.connection.is_disconnection_error(exc))
        else:
            self.fail("Disconnection was not caught.")
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")

    def test_connection_stays_disconnected_in_transaction(self):
        """Test that connection does not immediately reconnect."""
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")

    def test_reconnect_after_rollback(self):
        """Test that we reconnect after rolling back the connection."""
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.restart()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")
        self.connection.rollback()
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())

    def test_catch_disconnect_on_reconnect(self):
        """Test that reconnection failures result in DisconnectionError."""
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.stop()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")
        # Rollback the connection, but because the proxy is still
        # down, we get a DisconnectionError again.
        self.connection.rollback()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")

    def test_close_connection_after_disconnect(self):
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
        self.proxy.stop()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")
        self.connection.close()


class TwoPhaseCommitDisconnectionTest:

    def test_begin_after_rollback_with_disconnection_error(self):
        """
        If a rollback fails because of a disconnection error, the two-phase
        transaction should be properly reset.
        """
        xid1 = Xid(0, "foo", "bar")
        self.connection.begin(xid1)
        self.connection.execute("SELECT 1")
        self.proxy.stop()
        self.connection.rollback()
        self.proxy.start()
        xid2 = Xid(0, "egg", "baz")
        self.connection.begin(xid2)
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())

    def test_begin_after_with_statement_disconnection_error_and_rollback(self):
        """
        The two-phase transaction state is properly reset if a disconnection
        happens before the rollback.
        """
        xid1 = Xid(0, "foo", "bar")
        self.connection.begin(xid1)
        self.proxy.close()
        self.assertRaises(DisconnectionError,
                          self.connection.execute, "SELECT 1")
        self.connection.rollback()
        self.proxy.start()
        xid2 = Xid(0, "egg", "baz")
        self.connection.begin(xid2)
        result = self.connection.execute("SELECT 1")
        self.assertTrue(result.get_one())
